# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).




# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Tasks for running linting and formatting."""

from __future__ import annotations

from invoke.config import Config
from invoke.tasks import task


@task
def install(c: Config) -> None:
    """Install package with core and dev dependencies."""
    c.run("poetry install --sync --only base,main,lint")


@task
def check(c: Config) -> None:
    """Lint Python files."""
    commands: list[str] = [
        "poetry run ruff check .",
        "poetry run ruff format --check .",
        "poetry run pydoclint .",
    ]
    for command in commands:
        c.run(command)


@task(name="format")
def format_(c: Config) -> None:
    """Format Python files."""
    commands: list[str] = [
        "poetry run ruff --fix .",
        "poetry run ruff format .",
    ]
    for command in commands:
        c.run(command)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Tasks for generating Sphinx documentation."""

from invoke.config import Config
from invoke.tasks import task


@task
def install(c: Config) -> None:
    """Install package with core and docs dependencies."""
    c.run("poetry install --sync --only base,main,docs")


@task
def build(c: Config, *, watch: bool = True) -> None:
    """Build package Sphinx documentation."""
    if watch:
        command = (
            "poetry run sphinx-autobuild "
            "docs/source/ docs/build/html/ "
            "--watch docs/source/ --watch feud/ "
            "--ignore feud/_internal/"
        )
    else:
        command = "cd docs && make html"
    c.run(command)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

from collections import deque

import click
import pytest

from feud import typing as t
from feud.config import Config


def annotate(hint: t.Any) -> t.Annotated[t.Any, "annotation"]:
    return t.Annotated[hint, "annotation"]


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (bool, click.BOOL),
        (bool | None, click.BOOL),
        (annotate(bool) | None, click.BOOL),
        (int, click.INT),
        (int | None, click.INT),
        (annotate(int) | None, click.INT),
        (float, click.FLOAT),
        (float | None, click.FLOAT),
        (annotate(float) | None, click.FLOAT),
        (str, click.STRING),
        (str | None, click.STRING),
        (annotate(str) | None, click.STRING),
        (tuple, None),
        (tuple[int], (click.INT,)),
        (tuple[annotate(int)], (click.INT,)),
        (tuple[int, str], (click.INT, click.STRING)),
        (tuple[annotate(int), annotate(str)], (click.INT, click.STRING)),
        (tuple[int, ...], click.INT),
        (tuple[annotate(int), ...], click.INT),
        (tuple | None, None),
        (annotate(tuple) | None, None),
        (tuple[int] | None, (click.INT,)),
        (tuple[annotate(int)] | None, (click.INT,)),
        (tuple[int, str] | None, (click.INT, click.STRING)),
        (
            tuple[annotate(int), annotate(str)] | None,
            (click.INT, click.STRING),
        ),
        (tuple[int, ...] | None, click.INT),
        (tuple[annotate(int), ...] | None, click.INT),
        (list, None),
        (list[int], click.INT),
        (list[annotate(int)], click.INT),
        (list | None, None),
        (annotate(list) | None, None),
        (list[int] | None, click.INT),
        (list[annotate(int)] | None, click.INT),
        (set, None),
        (set[int], click.INT),
        (set[annotate(int)], click.INT),
        (set | None, None),
        (annotate(set) | None, None),
        (set[int] | None, click.INT),
        (set[annotate(int)] | None, click.INT),
        (frozenset, None),
        (frozenset[int], click.INT),
        (frozenset[annotate(int)], click.INT),
        (frozenset | None, None),
        (annotate(frozenset) | None, None),
        (frozenset[int] | None, click.INT),
        (frozenset[annotate(int)] | None, click.INT),
        (deque, None),
        (deque[int], click.INT),
        (deque[annotate(int)], click.INT),
        (deque | None, None),
        (annotate(deque) | None, None),
        (deque[int] | None, click.INT),
        (deque[annotate(int)] | None, click.INT),
        (type(None), None),
        (annotate(type(None)) | None, None),
    ],
)
def test_literal(
    helpers: type,
    *,
    config: Config,
    annotated: bool,
    hint: t.Any,
    expected: click.ParamType | None,
) -> None:
    helpers.check_get_click_type(
        config=config,
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import typing as t
from collections import OrderedDict
from operator import itemgetter
from pathlib import Path

import pytest

import feud
from feud import click


def assert_help(
    __obj: t.Union[
        feud.Group,
        click.Group,
        click.Command,
        t.Callable,
    ],
    /,
    *,
    capsys: pytest.CaptureFixture,
    expected: str,
) -> None:
    with pytest.raises(SystemExit):
        feud.run(__obj, ["--help"])
    out, _ = capsys.readouterr()
    assert out.strip() == expected.strip()


def test_undecorated_commands() -> None:
    class Test(feud.Group):
        def f(*, arg1: int) -> None:
            pass

        def g(*, arg1: int, arg2: bool) -> None:
            pass

    assert isinstance(Test.f, click.Command)
    assert len(Test.f.params) == 1

    assert isinstance(Test.g, click.Command)
    assert len(Test.g.params) == 2


def test_decorated_commands() -> None:
    class Test(feud.Group):
        @feud.command
        def f(*, arg1: int) -> None:
            pass

        @feud.command
        def g(*, arg1: int, arg2: bool) -> None:
            pass

    assert isinstance(Test.f, click.Command)
    assert len(Test.f.params) == 1

    assert isinstance(Test.g, click.Command)
    assert len(Test.g.params) == 2


def test_compile() -> None:
    class Test(feud.Group):
        def f(*, arg1: int) -> None:
            pass

        def g(*, arg1: int, arg2: bool) -> None:
            pass

    group = feud.compile(Test)
    assert isinstance(group, click.Group)

    assert len(group.commands) == 2
    assert list(group.commands.keys()) == ["f", "g"]


def test_run_group() -> None:
    class Test(feud.Group):
        def f(*, arg1: int) -> int:
            return arg1

    assert feud.run(Test, ["f", "--arg1", "2"], standalone_mode=False) == 2
    assert feud.run(Test.f, ["--arg1", "2"], standalone_mode=False) == 2


def test_call_group() -> None:
    class Test(feud.Group):
        def f(*, arg1: int) -> int:
            return arg1

    assert Test(["f", "--arg1", "2"], standalone_mode=False) == 2


def test_help_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(
        feud.Group,
        epilog="Visit https://www.com for more information.",
    ):
        """Useful commands."""

        def f(*, arg1: int) -> None:
            """Do something."""

        def g(*, arg1: int) -> None:
            """Perform something."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  Useful commands.

Options:
  --help  Show this message and exit.

Commands:
  f  Do something.
  g  Perform something.

  Visit https://www.com for more information.
        """,
    )


def test_help_click_kwarg(capsys: pytest.CaptureFixture) -> None:
    class Test(
        feud.Group,
        help="Useful commands.",
        epilog="Visit https://www.com for more information.",
    ):
        def f(*, arg1: int) -> None:
            """Do something."""

        def g(*, arg1: int) -> None:
            """Perform something."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  Useful commands.

Options:
  --help  Show this message and exit.

Commands:
  f  Do something.
  g  Perform something.

  Visit https://www.com for more information.
        """,
    )


def test_config_kwarg_propagation() -> None:
    class Test(feud.Group, show_help_defaults=False):
        def f(*, arg1: int = 1) -> None:
            pass

        def g(*, arg1: int = 2, arg2: bool = False) -> None:
            pass

    group = feud.compile(Test)
    for command in group.commands.values():
        for param in command.params:
            assert param.show_default is False

    class Test(feud.Group, show_help_defaults=True):
        def f(*, arg1: int = 1) -> None:
            pass

        def g(*, arg1: int = 2, arg2: bool = False) -> None:
            pass

    group = feud.compile(Test)
    for command in group.commands.values():
        for param in command.params:
            assert param.show_default is True


def test_config_propagation() -> None:
    class Test(feud.Group, config=feud.config(show_help_defaults=False)):
        def f(*, arg1: int = 1) -> None:
            pass

        def g(*, arg1: int = 2, arg2: bool = False) -> None:
            pass

    group = feud.compile(Test)
    for command in group.commands.values():
        for param in command.params:
            assert param.show_default is False

    class Test(feud.Group, config=feud.config(show_help_defaults=True)):
        def f(*, arg1: int = 1) -> None:
            pass

        def g(*, arg1: int = 2, arg2: bool = False) -> None:
            pass

    group = feud.compile(Test)
    for command in group.commands.values():
        for param in command.params:
            assert param.show_default is True


def test_config_kwarg_override() -> None:
    class Test(feud.Group, show_help_defaults=False):
        @feud.command(show_help_defaults=True)
        def f(*, arg1: int = 1) -> None:
            pass

    group = feud.compile(Test)
    for command in group.commands.values():
        for param in command.params:
            assert param.show_default is True


def test_subgroups_parent_single_child() -> None:
    class Parent(feud.Group):
        """This is the parent group."""

        def f(*, arg: int) -> None:
            """This is a command in the parent group."""
            return arg

    class Child(feud.Group):
        """This is a subgroup."""

        def g(*, arg: int) -> None:
            """This is a command in the subgroup."""
            return arg

    Parent.register(Child)
    assert Parent.subgroups() == [Child]
    assert Child.subgroups() == []

    parent = feud.compile(Parent)
    f, child = itemgetter("f", "child")(parent.commands)

    assert isinstance(f, click.Command)
    assert f.help == "This is a command in the parent group."

    assert isinstance(child, click.Group)
    assert child.help == "This is a subgroup."

    g = child.commands["g"]
    assert isinstance(g, click.Command)
    assert g.help == "This is a command in the subgroup."
    assert parent(["child", "g", "--arg", "1"], standalone_mode=False) == 1


def test_subgroups_parent_multi_children() -> None:
    """Parent subgroup with multiple children.

    Parent
    /      \
    Child1  Child2
    """

    class Parent(feud.Group):
        """This is the parent group."""

        def f(*, arg: int) -> int:
            """This is a command in the parent group."""
            return arg

    class Child1(feud.Group):
        """This is the first subgroup."""

        def g(*, arg: int) -> int:
            """This is a command in the first subgroup."""
            return arg

    class Child2(feud.Group):
        """This is the second subgroup."""

        def h(*, arg: int) -> int:
            """This is a command in the second subgroup."""
            return arg

    Parent.register([Child1, Child2])
    assert Parent.subgroups() == [Child1, Child2]
    assert Child1.subgroups() == []
    assert Child2.subgroups() == []

    parent = feud.compile(Parent)
    f, child1, child2 = itemgetter("f", "child1", "child2")(parent.commands)

    assert isinstance(f, click.Command)
    assert f.help == "This is a command in the parent group."
    assert parent(["f", "--arg", "1"], standalone_mode=False) == 1

    assert isinstance(child1, click.Group)
    assert child1.help == "This is the first subgroup."
    assert isinstance(child1.commands["g"], click.Command)
    assert (
        child1.commands["g"].help == "This is a command in the first subgroup."
    )
    assert parent(["child1", "g", "--arg", "1"], standalone_mode=False) == 1

    assert isinstance(child2, click.Group)
    assert child2.help == "This is the second subgroup."
    assert isinstance(child2.commands["h"], click.Command)
    assert (
        child2.commands["h"].help
        == "This is a command in the second subgroup."
    )
    assert parent(["child2", "h", "--arg", "1"], standalone_mode=False) == 1


def test_subgroups_nested() -> None:
    r"""Parent subgroup with multiple children.

    Parent
    /      \
        Child1  Child2
    /      \       \
    Child2  Child3  Child4
    \
         Child4
    """

    class Parent(feud.Group):
        """This is the parent group."""

        def f(*, arg: int) -> int:
            """This is a command in the parent group."""
            return arg

    class Child1(feud.Group):
        """This is the first subgroup."""

        def g(*, arg: int) -> int:
            """This is a command in the first subgroup."""
            return arg

    class Child2(feud.Group):
        """This is the second subgroup."""

        def h(*, arg: int) -> int:
            """This is a command in the second subgroup."""
            return arg

    class Child3(feud.Group):
        """This is the third subgroup."""

        def i(*, arg: int) -> int:
            """This is a command in the third subgroup."""
            return arg

    class Child4(feud.Group):
        """This is the fourth subgroup."""

        def j(*, arg: int) -> int:
            """This is a command in the fourth subgroup."""
            return arg

    Parent.register(Child1)
    Parent.register(Child2)
    Child1.register(Child2)
    Child1.register(Child3)
    Child2.register(Child4)

    assert Parent.subgroups() == [Child1, Child2]
    assert Child1.subgroups() == [Child2, Child3]
    assert Child2.subgroups() == [Child4]
    assert Child3.subgroups() == []
    assert Child4.subgroups() == []

    Parent.deregister()
    Child1.deregister()
    Child2.deregister()
    Child3.deregister()
    Child4.deregister()

    Parent.register([Child1, Child2])
    Child1.register([Child2, Child3])
    Child2.register(Child4)

    assert Parent.subgroups() == [Child1, Child2]
    assert Child1.subgroups() == [Child2, Child3]
    assert Child2.subgroups() == [Child4]
    assert Child3.subgroups() == []
    assert Child4.subgroups() == []

    Parent.deregister()
    Child1.deregister()
    Child2.deregister()
    Child3.deregister()
    Child4.deregister()

    Parent.register([Child1, Child2])
    Child1.register([Child2, Child3])
    Child2.register(Child4)

    assert Parent.subgroups() == [Child1, Child2]
    assert Child1.subgroups() == [Child2, Child3]
    assert Child2.subgroups() == [Child4]
    assert Child3.subgroups() == []
    assert Child4.subgroups() == []

    parent = feud.compile(Parent)

    assert len(parent.commands) == 3
    f, child1, child2 = itemgetter("f", "child1", "child2")(parent.commands)
    assert isinstance(f, click.Command)
    assert isinstance(child1, click.Group)
    assert isinstance(child2, click.Group)

    assert f.help == "This is a command in the parent group."
    assert parent(["f", "--arg", "1"], standalone_mode=False) == 1

    assert len(child2.commands) == 2
    h, child4 = itemgetter("h", "child4")(child2.commands)
    assert isinstance(h, click.Command)
    assert isinstance(child4, click.Group)

    assert h.help == "This is a command in the second subgroup."
    assert parent(["child2", "h", "--arg", "1"], standalone_mode=False) == 1

    assert len(child4.commands) == 1
    j = child4.commands["j"]
    assert isinstance(j, click.Command)
    assert j.help == "This is a command in the fourth subgroup."
    assert (
        parent(["child2", "child4", "j", "--arg", "1"], standalone_mode=False)
        == 1
    )

    assert len(child1.commands) == 3
    g, child2, child3 = itemgetter("g", "child2", "child3")(child1.commands)
    assert isinstance(g, click.Command)
    assert isinstance(child2, click.Group)
    assert isinstance(child3, click.Group)

    assert g.help == "This is a command in the first subgroup."
    assert parent(["child1", "g", "--arg", "1"], standalone_mode=False) == 1

    assert len(child2.commands) == 2
    h, child4 = itemgetter("h", "child4")(child2.commands)
    assert isinstance(h, click.Command)
    assert isinstance(child4, click.Group)

    assert h.help == "This is a command in the second subgroup."
    assert parent(["child2", "h", "--arg", "1"], standalone_mode=False) == 1

    assert len(child4.commands) == 1
    j = child4.commands["j"]
    assert isinstance(j, click.Command)
    assert j.help == "This is a command in the fourth subgroup."
    assert (
        parent(["child2", "child4", "j", "--arg", "1"], standalone_mode=False)
        == 1
    )

    assert len(child2.commands) == 2
    h = child2.commands["h"]
    assert isinstance(h, click.Command)
    assert h.help == "This is a command in the second subgroup."
    assert (
        parent(["child1", "child2", "h", "--arg", "1"], standalone_mode=False)
        == 1
    )

    assert len(child3.commands) == 1
    i = child3.commands["i"]
    assert isinstance(i, click.Command)
    assert i.help == "This is a command in the third subgroup."
    assert (
        parent(["child1", "child3", "i", "--arg", "1"], standalone_mode=False)
        == 1
    )


def test_deregister_nested() -> None:
    class Parent(feud.Group):
        """This is the parent group."""

        def f(*, arg: int) -> int:
            """This is a command in the parent group."""
            return arg

    class Child1(feud.Group):
        """This is the first subgroup."""

        def g(*, arg: int) -> int:
            """This is a command in the first subgroup."""
            return arg

    class Child2(feud.Group):
        """This is the second subgroup."""

        def h(*, arg: int) -> int:
            """This is a command in the second subgroup."""
            return arg

    class Child3(feud.Group):
        """This is the third subgroup."""

        def i(*, arg: int) -> int:
            """This is a command in the third subgroup."""
            return arg

    Parent.register([Child1, Child2])
    Child1.register(Child2)
    Child2.register(Child3)

    assert Parent.subgroups() == [Child1, Child2]
    assert Child1.subgroups() == [Child2]
    assert Child2.subgroups() == [Child3]

    parent = feud.compile(Parent)

    assert len(parent.commands) == 3
    f, child1, child2 = itemgetter("f", "child1", "child2")(parent.commands)
    assert isinstance(f, click.Command)
    assert isinstance(child1, click.Group)
    assert isinstance(child2, click.Group)

    assert f.help == "This is a command in the parent group."
    assert parent(["f", "--arg", "1"], standalone_mode=False) == 1

    assert len(child2.commands) == 2
    h, child3 = itemgetter("h", "child3")(child2.commands)
    assert isinstance(h, click.Command)
    assert isinstance(child3, click.Group)

    assert h.help == "This is a command in the second subgroup."
    assert parent(["child2", "h", "--arg", "1"], standalone_mode=False) == 1

    assert len(child3.commands) == 1
    i = child3.commands["i"]
    assert isinstance(i, click.Command)

    assert i.help == "This is a command in the third subgroup."
    assert (
        parent(["child2", "child3", "i", "--arg", "1"], standalone_mode=False)
        == 1
    )

    assert len(child1.commands) == 2
    g, child2 = itemgetter("g", "child2")(child1.commands)
    assert isinstance(g, click.Command)
    assert isinstance(child2, click.Group)

    assert g.help == "This is a command in the first subgroup."
    assert parent(["child1", "g", "--arg", "1"], standalone_mode=False) == 1

    assert len(child2.commands) == 2
    h, child3 = itemgetter("h", "child3")(child2.commands)
    assert isinstance(h, click.Command)
    assert isinstance(child3, click.Group)

    assert h.help == "This is a command in the second subgroup."
    assert (
        parent(["child1", "child2", "h", "--arg", "1"], standalone_mode=False)
        == 1
    )

    Child2.deregister(Child3)

    assert Parent.subgroups() == [Child1, Child2]
    assert Child1.subgroups() == [Child2]
    assert Child2.subgroups() == []

    parent = feud.compile(Parent)

    assert len(parent.commands) == 3
    f, child1, child2 = itemgetter("f", "child1", "child2")(parent.commands)
    assert isinstance(f, click.Command)
    assert isinstance(child1, click.Group)
    assert isinstance(child2, click.Group)

    assert f.help == "This is a command in the parent group."
    assert parent(["f", "--arg", "1"], standalone_mode=False) == 1

    assert len(child2.commands) == 1
    h = child2.commands["h"]
    assert isinstance(h, click.Command)

    assert h.help == "This is a command in the second subgroup."
    assert parent(["child2", "h", "--arg", "1"], standalone_mode=False) == 1

    assert len(child1.commands) == 2
    g, child2 = itemgetter("g", "child2")(child1.commands)
    assert isinstance(g, click.Command)
    assert isinstance(child2, click.Group)

    assert g.help == "This is a command in the first subgroup."
    assert parent(["child1", "g", "--arg", "1"], standalone_mode=False) == 1

    assert len(child2.commands) == 1
    h = child2.commands["h"]
    assert isinstance(child3, click.Group)

    assert h.help == "This is a command in the second subgroup."
    assert (
        parent(["child1", "child2", "h", "--arg", "1"], standalone_mode=False)
        == 1
    )


def test_inheritance_config_kwargs() -> None:
    class Parent(
        feud.Group,
        negate_flags=False,
        show_help_defaults=False,
    ):
        pass

    class Child(Parent):
        pass

    assert Child.__feud_config__ == feud.config(
        negate_flags=False, show_help_defaults=False
    )

    class Child(Parent, show_help_defaults=True):
        pass

    assert Child.__feud_config__ == feud.config(
        negate_flags=False, show_help_defaults=True
    )


def test_inheritance_config_class() -> None:
    class Parent(
        feud.Group,
        config=feud.config(negate_flags=False, show_help_defaults=False),
    ):
        pass

    class Child(Parent):
        pass

    assert Child.__feud_config__ == feud.config(
        negate_flags=False, show_help_defaults=False
    )

    class Child(Parent, show_help_defaults=True):
        pass

    assert Child.__feud_config__ == feud.config(
        negate_flags=False, show_help_defaults=True
    )

    class Child(
        Parent,
        config=feud.config(show_help_defaults=True),
        show_help_datetime_formats=True,
    ):
        pass

    assert Child.__feud_config__ == feud.config(
        negate_flags=False,
        show_help_defaults=True,
        show_help_datetime_formats=True,
    )


def test_inheritance_ancestor() -> None:
    class Grandparent(feud.Group, name="older", negate_flags=False):
        def f(*, arg: int) -> int:
            return arg

    grandparent = feud.compile(Grandparent)
    assert grandparent.name == "older"
    assert list(grandparent.commands) == ["f"]

    class Parent(Grandparent, name="old", show_help_defaults=False):
        def g(*, arg: int) -> int:
            return arg

    parent = feud.compile(Parent)
    assert parent.name == "old"
    assert list(parent.commands) == ["f", "g"]
    assert all(
        isinstance(command, click.Command)
        for command in parent.commands.values()
    )

    class Child(Parent, name="young", show_help_datetime_formats=True):
        def h(*, arg: int) -> int:
            return arg

    child = feud.compile(Child)
    assert child.name == "young"
    assert list(child.commands) == ["f", "g", "h"]
    assert all(
        isinstance(command, click.Command)
        for command in child.commands.values()
    )
    assert Child.__feud_config__ == feud.config(
        negate_flags=False,
        show_help_defaults=False,
        show_help_datetime_formats=True,
    )


def test_inheritance_multiple() -> None:
    class Mother(
        feud.Group, negate_flags=False, epilog="Are we one big family?"
    ):
        def f(*, arg: int) -> int:
            return arg

    mother = feud.compile(Mother)
    assert mother.name == "mother"
    assert list(mother.commands) == ["f"]

    class Father(
        feud.Group, show_help_defaults=False, epilog="We are one big family!"
    ):
        def g(*, arg: int) -> int:
            return arg

    father = feud.compile(Father)
    assert father.name == "father"
    assert list(father.commands) == ["g"]

    class Child(Mother, Father, show_help_datetime_formats=True):
        def h(*, arg: int) -> int:
            return arg

    child = feud.compile(Child)
    assert list(child.commands) == ["f", "g", "h"]
    assert Child.__feud_config__ == feud.config(
        negate_flags=False,
        show_help_defaults=False,
        show_help_datetime_formats=True,
    )
    assert Child.__feud_click_kwargs__ == {
        "name": "father",
        "epilog": "We are one big family!",
    }

    class Child(Father, Mother, show_help_datetime_formats=True):
        def h(*, arg: int) -> int:
            return arg

    child = feud.compile(Child)
    assert list(child.commands) == ["g", "f", "h"]
    assert Child.__feud_config__ == feud.config(
        negate_flags=False,
        show_help_defaults=False,
        show_help_datetime_formats=True,
    )
    assert Child.__feud_click_kwargs__ == {
        "name": "mother",
        "epilog": "Are we one big family?",
    }


def test_inheritance_overwrite_command() -> None:
    class Parent(feud.Group):
        def f(*, arg: str = "From the parent!") -> int:
            return arg

    class Child(Parent):
        def f(*, arg: str = "From the child!") -> str:
            return arg

    assert Child.f([], standalone_mode=False) == "From the child!"


def test_register_deregister_compile() -> None:
    class Parent(
        feud.Group,
        negate_flags=False,
        show_help_defaults=False,
        epilog="Visit https://www.com for more information.",
    ):
        """This is the parent group."""

        def f(*, arg: int) -> int:
            """This is a command in the parent group."""
            return arg

    class Subgroup(feud.Group):
        def g(*, arg: int) -> int:
            """This is a command in a subgroup."""
            return arg

    Parent.register(Subgroup)

    class Child(Parent):
        pass

    child = feud.compile(Child)
    assert list(child.commands) == ["f", "subgroup"]

    Child.deregister(Subgroup)

    child = feud.compile(Child)
    assert list(child.commands) == ["f"]

    class Child(Parent):
        pass

    parent = feud.compile(Parent)
    assert list(parent.commands) == ["f", "subgroup"]

    child = feud.compile(Child)
    assert list(child.commands) == ["f", "subgroup"]

    Parent.deregister(Subgroup)

    parent = feud.compile(Parent)
    assert list(parent.commands) == ["f"]

    child = feud.compile(Child)
    assert list(child.commands) == ["f", "subgroup"]


def test_register() -> None:
    class A(feud.Group):
        pass

    class B(feud.Group):
        pass

    class C(feud.Group):
        pass

    # A -> B - should allow
    A.register(B)
    assert A.subgroups() == [B]
    A.deregister()
    assert A.subgroups() == []

    # A -> B and A -> B - should allow but with a warning
    msg = (
        f"Group {B.__name__!r} is already registered as a "
        f"subgroup under {A.__name__!r} and will be ignored."
    )
    with pytest.warns(RuntimeWarning, match=msg):
        A.register([B, B])
    assert A.subgroups() == [B]
    A.deregister()
    assert A.subgroups() == []

    # A -> B and A -> C - should allow
    A.register([B, C])
    assert A.subgroups() == [B, C]
    A.deregister()
    assert A.subgroups() == []

    # A -> B, B -> A - shouldn't allow (circular dependency)
    A.register(B)
    msg = (
        f"Group {B.__name__!r} is a descendant subgroup of {A.__name__!r}, "
        "causing a circular dependency."
    )
    with pytest.raises(feud.RegistrationError, match=msg):
        B.register(A)
    assert A.subgroups() == [B]
    assert B.subgroups() == []
    for group in (A, B):
        group.deregister()
        assert group.subgroups() == []

    # A -> A - shouldn't allow (circular dependency)
    msg = f"Group {A.__name__!r} cannot be a subgroup of itself."
    with pytest.raises(feud.RegistrationError, match=msg):
        A.register(A)
    assert A.subgroups() == []
    A.deregister()
    assert A.subgroups() == []

    # A -> B, B -> C, C -> A - shouldn't allow (circular dependency)
    A.register(B)
    B.register(C)
    msg = (
        f"Group {C.__name__!r} is a descendant subgroup of {A.__name__!r}, "
        "causing a circular dependency."
    )
    with pytest.raises(feud.RegistrationError, match=msg):
        C.register(A)
    assert A.subgroups() == [B]
    assert B.subgroups() == [C]
    assert C.subgroups() == []
    for group in (A, B, C):
        group.deregister()
        assert group.subgroups() == []


def test_deregister() -> None:
    class A(feud.Group):
        pass

    class B(feud.Group):
        pass

    class C(feud.Group):
        pass

    A.register(B)
    assert A.subgroups() == [B]
    A.deregister(B)
    assert A.subgroups() == []

    msg = (
        f"Group {A.__name__!r} is not a registered subgroup under "
        f"{A.__name__!r} and will be ignored."
    )
    with pytest.warns(RuntimeWarning, match=msg):
        A.deregister(A)
    assert A.subgroups() == []

    msg = (
        f"Group {B.__name__!r} is not a registered subgroup under "
        f"{A.__name__!r} and will be ignored."
    )
    with pytest.warns(RuntimeWarning, match=msg):
        A.deregister(B)
    assert A.subgroups() == []

    A.register([B, C])
    assert A.subgroups() == [B, C]
    A.deregister(B)
    assert A.subgroups() == [C]
    A.deregister(C)
    assert A.subgroups() == []

    A.register([B, C])
    assert A.subgroups() == [B, C]
    A.deregister([B, C])
    assert A.subgroups() == []

    A.register([B, C])
    assert A.subgroups() == [B, C]
    msg = (
        f"Group {A.__name__!r} is not a registered subgroup under "
        f"{A.__name__!r} and will be ignored."
    )
    with pytest.warns(RuntimeWarning, match=msg):
        A.deregister(A)
    assert A.subgroups() == [B, C]
    A.deregister()

    A.register(B)
    msg = (
        f"Group {B.__name__!r} is not a registered subgroup under "
        f"{A.__name__!r} and will be ignored."
    )
    with pytest.warns(RuntimeWarning, match=msg):
        A.deregister([B, B])
    assert A.subgroups() == []


def test_descendants() -> None:
    class A(feud.Group):
        pass

    class B(feud.Group):
        pass

    class C(feud.Group):
        pass

    class D(feud.Group):
        pass

    class E(feud.Group):
        pass

    class F(feud.Group):
        pass

    A.register([B, C, D])
    B.register(D)
    C.register(E)
    E.register(F)

    assert F.descendants() == OrderedDict()
    assert E.descendants() == OrderedDict([(F, F.descendants())])
    assert C.descendants() == OrderedDict([(E, E.descendants())])
    assert D.descendants() == OrderedDict()
    assert B.descendants() == OrderedDict([(D, D.descendants())])
    assert A.descendants() == OrderedDict(
        [
            (B, B.descendants()),
            (C, C.descendants()),
            (D, D.descendants()),
        ]
    )


def test_help_no_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        pass

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.
        """,
    )


def test_help_simple_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        """This is a group."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This is a group.

Options:
  --help  Show this message and exit.
        """,
    )


def test_help_param_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        """This is a group.\f

        Parameters
        ----------
        param:
            Help for a parameter.
        """

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This is a group.

Options:
  --help  Show this message and exit.
        """,
    )


def test_help_override(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group, help="Overridden."):
        """This is a group."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  Overridden.

Options:
  --help  Show this message and exit.
        """,
    )


def test_main_help_no_docstrings(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        def __main__() -> None:
            pass

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.
        """,
    )


def test_main_help_class_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        """This is a class-level docstring."""

        def __main__() -> None:
            pass

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This is a class-level docstring.

Options:
  --help  Show this message and exit.
        """,
    )


def test_main_help_function_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        def __main__() -> None:
            """This is a function-level docstring."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This is a function-level docstring.

Options:
  --help  Show this message and exit.
        """,
    )


def test_main_help_both_docstrings(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        """This is a class-level docstring."""

        def __main__() -> None:
            """This is a function-level docstring."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This is a function-level docstring.

Options:
  --help  Show this message and exit.
        """,
    )


def test_main_help_both_docstrings_with_override(
    capsys: pytest.CaptureFixture
) -> None:
    class Test(feud.Group, help="Overridden."):
        """This is a class-level docstring."""

        def __main__() -> None:
            """This is a function-level docstring."""

    assert_help(
        Test,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  Overridden.

Options:
  --help  Show this message and exit.
        """,
    )


def test_main(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group, invoke_without_command=True):
        """This group does something relative to a root directory.\f

        Parameters
        ----------
        root:
            Root directory
        """

        @staticmethod
        @click.version_option("0.1.0")
        @feud.alias(root="-r")
        def __main__(ctx: click.Context, *, root: Path = Path(".")) -> None:
            ctx.obj = {"root": root}
            return root

        @staticmethod
        def command(ctx: click.Context, path: Path) -> Path:
            """Returns a full path.\f

            Parameters
            ----------
            path:
                Relative path.
            """
            return ctx.obj["root"] / path

    group = feud.compile(Test)

    assert_help(
        group,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This group does something relative to a root directory.

Options:
  -r, --root PATH  Root directory  [default: .]
  --version        Show the version and exit.
  --help           Show this message and exit.

Commands:
  command  Returns a full path.
        """,
    )

    # check version
    with pytest.raises(SystemExit):
        group(["--version"])
    out, _ = capsys.readouterr()
    assert out.strip() == "pytest, version 0.1.0"

    # test invoke without command
    assert group(["-r", "/usr"], standalone_mode=False) == Path("/usr")

    # test command context
    assert group(
        ["-r", "/usr", "command", "bin/sh"],
        standalone_mode=False,
    ) == Path("/usr/bin/sh")


def test_main_inheritance_no_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group, invoke_without_command=True):
        """This group does something relative to a root directory.\f

        Parameters
        ----------
        root:
            Root directory
        """

        @staticmethod
        @click.version_option("0.1.0")
        @feud.alias(root="-r")
        def __main__(ctx: click.Context, *, root: Path = Path(".")) -> None:
            ctx.obj = {"root": root}
            return root

        @staticmethod
        def command(ctx: click.Context, path: Path) -> Path:
            """Returns a full path.\f

            Parameters
            ----------
            path:
                Relative path.
            """
            return ctx.obj["root"] / path

    class Child(Test):
        pass

    group = feud.compile(Child)

    assert_help(
        group,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This group does something relative to a root directory.

Options:
  -r, --root PATH  Root directory  [default: .]
  --version        Show the version and exit.
  --help           Show this message and exit.

Commands:
  command  Returns a full path.
        """,
    )

    # check version
    with pytest.raises(SystemExit):
        group(["--version"])
    out, _ = capsys.readouterr()
    assert out.strip() == "pytest, version 0.1.0"

    # test invoke without command
    assert group(["-r", "/usr"], standalone_mode=False) == Path("/usr")

    # test command context
    assert group(
        ["-r", "/usr", "command", "bin/sh"],
        standalone_mode=False,
    ) == Path("/usr/bin/sh")


def test_main_inheritance_docstring(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group, invoke_without_command=True):
        """This group does something relative to a root directory.\f

        Parameters
        ----------
        root:
            Root directory
        """

        @staticmethod
        @click.version_option("0.1.0")
        @feud.alias(root="-r")
        def __main__(ctx: click.Context, *, root: Path = Path(".")) -> None:
            ctx.obj = {"root": root}
            return root

        @staticmethod
        def command(ctx: click.Context, path: Path) -> Path:
            """Returns a full path.\f

            Parameters
            ----------
            path:
                Relative path.
            """
            return ctx.obj["root"] / path

    class Child(Test):
        """This is a new docstring.\f

        Parameters
        ----------
        root:
            Root directory
        """

    group = feud.compile(Child)

    assert_help(
        group,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] COMMAND [ARGS]...

  This is a new docstring.

Options:
  -r, --root PATH  Root directory  [default: .]
  --version        Show the version and exit.
  --help           Show this message and exit.

Commands:
  command  Returns a full path.
        """,
    )

    # check version
    with pytest.raises(SystemExit):
        group(["--version"])
    out, _ = capsys.readouterr()
    assert out.strip() == "pytest, version 0.1.0"

    # test invoke without command
    assert group(["-r", "/usr"], standalone_mode=False) == Path("/usr")

    # test command context
    assert group(
        ["-r", "/usr", "command", "bin/sh"],
        standalone_mode=False,
    ) == Path("/usr/bin/sh")



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import enum

import click
import pytest

from feud import typing as t
from feud._internal import _types
from feud.config import Config


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        *[
            (
                path_type,
                lambda x: isinstance(x, click.Path) and x.exists is False,
            )
            for path_type in _types.click.PATH_TYPES
        ],
        (t.UUID, click.UUID),
        (t.Decimal, click.FLOAT),
        (
            t.date,
            lambda x: isinstance(x, _types.click.DateTime)
            and x.name == t.date.__name__,
        ),
        (
            t.time,
            lambda x: isinstance(x, _types.click.DateTime)
            and x.name == t.time.__name__,
        ),
        (
            t.datetime,
            lambda x: isinstance(x, _types.click.DateTime)
            and x.name == t.datetime.__name__,
        ),
        (
            t.timedelta,
            lambda x: isinstance(x, _types.click.DateTime)
            and x.name == t.timedelta.__name__,
        ),
        (
            enum.Enum("TestEnum", {"A": "a", "B": "b"}),
            lambda x: isinstance(x, click.Choice) and x.choices == ["a", "b"],
        ),
    ],
)
def test_stdlib(
    helpers: type,
    *,
    config: Config,
    annotated: bool,
    hint: t.Any,
    expected: click.ParamType | None,
) -> None:
    helpers.check_get_click_type(
        config=config,
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import click
import pydantic as pyd
import pytest

from feud import typing as t
from feud._internal._types.click import DateTime
from feud.config import Config


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        # pydantic types
        (t.AmqpDsn, None),
        (t.AnyHttpUrl, None),
        (t.AnyUrl, None),
        (
            t.AwareDatetime,
            lambda x: (
                isinstance(x, DateTime) and x.name == t.datetime.__name__
            ),
        ),
        (t.Base64Bytes, None),
        (t.Base64Str, click.STRING),
        (t.ByteSize, None),
        (t.CockroachDsn, None),
        (
            t.DirectoryPath,
            lambda x: isinstance(x, click.Path) and x.exists is True,
        ),
        (t.EmailStr, None),
        (
            t.FilePath,
            lambda x: isinstance(x, click.Path) and x.exists is True,
        ),
        (t.FileUrl, None),
        (t.FiniteFloat, click.FLOAT),
        (
            t.FutureDate,
            lambda x: isinstance(x, DateTime) and x.name == t.date.__name__,
        ),
        (
            t.FutureDatetime,
            lambda x: (
                isinstance(x, DateTime) and x.name == t.datetime.__name__
            ),
        ),
        (t.HttpUrl, None),
        (t.IPvAnyAddress, None),
        (t.IPvAnyInterface, None),
        (t.IPvAnyNetwork, None),
        (t.ImportString, None),
        (t.Json, None),
        (t.KafkaDsn, None),
        (t.MariaDBDsn, None),
        (t.MongoDsn, None),
        (t.MySQLDsn, None),
        (
            t.NaiveDatetime,
            lambda x: (
                isinstance(x, DateTime) and x.name == t.datetime.__name__
            ),
        ),
        (t.NameEmail, None),
        (
            t.NegativeFloat,
            lambda x: isinstance(x, click.FloatRange)
            and x.min is None
            and x.min_open is False
            and x.max == 0
            and x.max_open is True,
        ),
        (
            t.NegativeInt,
            lambda x: isinstance(x, click.IntRange)
            and x.min is None
            and x.min_open is False
            and x.max == 0
            and x.max_open is True,
        ),
        (
            t.NewPath,
            lambda x: isinstance(x, click.Path) and x.exists is False,
        ),
        (
            t.NonNegativeFloat,
            lambda x: isinstance(x, click.FloatRange)
            and x.min == 0
            and x.min_open is False
            and x.max is None
            and x.max_open is False,
        ),
        (
            t.NonNegativeInt,
            lambda x: isinstance(x, click.IntRange)
            and x.min == 0
            and x.min_open is False
            and x.max is None
            and x.max_open is False,
        ),
        (
            t.NonPositiveFloat,
            lambda x: isinstance(x, click.FloatRange)
            and x.min is None
            and x.min_open is False
            and x.max == 0
            and x.max_open is False,
        ),
        (
            t.NonPositiveInt,
            lambda x: isinstance(x, click.IntRange)
            and x.min is None
            and x.min_open is False
            and x.max == 0
            and x.max_open is False,
        ),
        (
            t.PastDate,
            lambda x: isinstance(x, DateTime) and x.name == t.date.__name__,
        ),
        (
            t.PastDatetime,
            lambda x: (
                isinstance(x, DateTime) and x.name == t.datetime.__name__
            ),
        ),
        (
            t.PositiveFloat,
            lambda x: isinstance(x, click.FloatRange)
            and x.min == 0
            and x.min_open is True
            and x.max is None
            and x.max_open is False,
        ),
        (
            t.PositiveInt,
            lambda x: isinstance(x, click.IntRange)
            and x.min == 0
            and x.min_open is True
            and x.max is None
            and x.max_open is False,
        ),
        (t.PostgresDsn, None),
        (t.RedisDsn, None),
        (t.SecretBytes, None),
        (t.SecretStr, None),
        (t.SkipValidation, None),
        (t.StrictBool, click.BOOL),
        (t.StrictBytes, None),
        (t.StrictFloat, click.FLOAT),
        (t.StrictInt, click.INT),
        (t.StrictStr, click.STRING),
        (t.UUID1, click.UUID),
        (t.UUID3, click.UUID),
        (t.UUID4, click.UUID),
        (t.UUID5, click.UUID),
        (t.conbytes(max_length=1), None),
        (
            t.condate(lt=t.date.today()),
            lambda x: isinstance(x, DateTime) and x.name == t.date.__name__,
        ),
        (
            t.condecimal(lt=t.Decimal("3.14"), ge=t.Decimal("0.01")),
            lambda x: isinstance(x, click.FloatRange)
            and x.min == t.Decimal("0.01")
            and x.min_open is False
            and x.max == t.Decimal("3.14")
            and x.max_open is True,
        ),
        (
            t.confloat(lt=3.14, ge=0.01),
            lambda x: isinstance(x, click.FloatRange)
            and x.min == 0.01
            and x.min_open is False
            and x.max == 3.14
            and x.max_open is True,
        ),
        (t.confrozenset(int, max_length=1), click.INT),
        (
            t.conint(lt=3, ge=0),
            lambda x: isinstance(x, click.IntRange)
            and x.min == 0
            and x.min_open is False
            and x.max == 3
            and x.max_open is True,
        ),
        (t.conlist(int, max_length=1), click.INT),
        (t.conset(int, max_length=1), click.INT),
        (t.constr(max_length=1), click.STRING),
        (t.Annotated[int, pyd.AfterValidator(lambda x: x + 1)], click.INT),
    ],
)
def test_pydantic(
    helpers: type,
    *,
    config: Config,
    annotated: bool,
    hint: t.Any,
    expected: click.ParamType | None,
) -> None:
    helpers.check_get_click_type(
        config=config,
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Version information for Feud.

Source code modified from pydantic (https://github.com/pydantic/pydantic).

    The MIT License (MIT)

    Copyright (c) 2017 to present Pydantic Services Inc. and individual
    contributors.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
"""

__all__ = ["VERSION", "version_info"]

VERSION = "0.1.5"


def version_info() -> str:
    """Return complete version information for Feud and its dependencies."""
    import importlib.metadata as importlib_metadata
    import platform
    import sys
    from pathlib import Path

    # get data about packages that:
    # - are closely related to feud,
    # - use feud,
    # - often conflict with feud.
    package_names = {
        "click",
        "pydantic",
        "docstring-parser",
        "rich-click",
        "rich",
        "email-validator",
        "pydantic-extra-types",
        "phonenumbers",
        "pycountry",
    }
    related_packages = []

    for dist in importlib_metadata.distributions():
        name = dist.metadata["Name"]
        if name in package_names:
            related_packages.append(f"{name}-{dist.version}")

    info = {
        "feud version": VERSION,
        "install path": Path(__file__).resolve().parent,
        "python version": sys.version,
        "platform": platform.platform(),
        "related packages": " ".join(related_packages),
    }
    return "\n".join(
        "{:>30} {}".format(k + ":", str(v).replace("\n", " "))
        for k, v in info.items()
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Build powerful CLIs with simple idiomatic Python, driven by type hints.
Not all arguments are bad.
"""

import feud.version

__version__ = feud.version.VERSION

from feud import click as click
from feud import exceptions as exceptions
from feud import typing as typing
from feud.config import *
from feud.core import *
from feud.decorators import *
from feud.exceptions import *



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Core utilities for building and running a command-line interface."""

from __future__ import annotations

import contextlib
import inspect
import sys
import types
import typing as t
import warnings

try:
    import rich_click as click

    RICH = True
except ImportError:
    import click

    RICH = False

import feud.exceptions
from feud.config import Config
from feud.core.command import *
from feud.core.group import *

__all__ = ["Group", "compile", "command", "run"]

RICH_DEFAULTS = {"SHOW_ARGUMENTS": True}

Runner = t.Union[
    click.Command,
    type[Group],
    t.Callable,
    types.ModuleType,
    # t.Iterable[click.Command | type[Group] | t.Callable | types.ModuleType],
    t.Iterable,
    # dict[str, click.Command | type[Group] | t.Callable | types.ModuleType],
    dict,
    # dict[str, "Runner"],
]


def run(
    obj: Runner | None = None,
    /,
    args: list[str] | None = None,
    *,
    name: str | None = None,
    help: str | None = None,  # noqa: A002
    epilog: str | None = None,
    config: Config | None = None,
    rich_settings: dict[str, t.Any] | None = None,
    warn: bool = True,
    **click_kwargs: t.Any,
) -> t.Any:
    """

    Run a function, :py:class:`click.Command`, :py:class:`.Group`,
    or :py:class:`click.Group`.

    Multiple functions/commands can also be provided as a :py:obj:`dict`,
    iterable or module object. :py:obj:`dict` objects may also be nested.

    If called on a function, it will be automatically decorated with
    :py:func:`.command` using the default configuration to convert it into a
    :py:class:`click.Command` which will then be executed.

    If no runnable object is provided, automatic discovery will be done
    on the current module.

    .. warning::

        ``name``, ``help``, ``epilog`` and ``config`` are ignored if a
        :py:class:`click.Command`, :py:class:`click.Group` or
        :py:class:`.Group` is provided.

    Parameters
    ----------
    obj:
        Runnable group, command or function to run, or :py:obj:`dict`,
        iterable or module of runnable objects.

    args:
        Command-line arguments provided to
        :py:class:`click.Command`.

    name:
        CLI command or group name.

    help:
        CLI command or group description, displayed when ``--help``
        is called. If not set, the docstring of the object will be used if
        available.

    epilog:
        CLI command or group epilog. Appears at the bottom of ``--help``.

    config:
        Configuration for the command or group.

        If a :py:obj:`dict`, iterable or module ``obj`` is provided, the
        configuration will be forwarded to the nested runnable objects within.

    rich_settings:
        Styling options for ``rich-click``, e.g.
        ``{"SHOW_ARGUMENTS": False, "MAX_WIDTH": 60}``.

        See all available options
        `here <https://github.com/ewels/rich-click/blob/e6a3add46c591d49079d440917700dfe28cf0cfe/src/rich_click/rich_click.py#L48-L131>`__
        (as of ``rich-click`` v1.7.2).

    warn:
        Silences warnings that are produced if ``name``, ``help``, ``epilog``
        or ``config`` are provided when ``obj`` is a :py:class:`click.Command`,
        :py:class:`click.Group` or :py:class:`.Group`.

    **click_kwargs:
        Additional keyword arguments provided to
        :py:class:`click.Command`.

    Returns
    -------
    Output of the called object.

    Raises
    ------
    feud.exceptions.CompilationError
        If no runnable object or current module can be determined.

    Examples
    --------
    Running an undecorated function.

    >>> import feud
    >>> def func(*, opt: int) -> int:
    ...     return opt
    >>> feud.run(func, ["--opt", "3"], standalone_mode=False)
    3

    Running a :py:class:`click.Command`.

    >>> import feud
    >>> @feud.command
    ... def func(*, opt: int) -> int:
    ...     return opt
    >>> feud.run(func, ["--opt", "3"], standalone_mode=False)
    3

    Running a :py:class:`.Group`.

    >>> import feud
    >>> class CLI(feud.Group):
    ...     def func(*, opt: int) -> int:
    ...         return opt
    >>> feud.run(CLI, ["func", "--opt", "3"], standalone_mode=False)
    3

    Running a :py:class:`click.Group`.

    >>> import feud
    >>> from feud import click
    >>> class CLI(feud.Group):
    ...     def func(*, opt: int) -> int:
    ...         return opt
    >>> group: click.Group = feud.compile(CLI)
    >>> feud.run(group, ["func", "--opt", "3"], standalone_mode=False)
    3

    Running a :py:obj:`dict` of functions, commands, groups or modules.

    >>> import feud
    >>> def func1(*, opt: int) -> int:
    ...     return opt
    >>> def func2(*, opt: float) -> float:
    ...     return opt
    >>> feud.run(
    ...     {"f": func1, "g": func2},
    ...     ["g", "--opt", "0.12"],
    ...     standalone_mode=False,
    ... )
    0.12

    Running an iterable of functions, commands, groups or modules.

    >>> import feud
    >>> def func1(*, opt: int) -> int:
    ...     return opt
    >>> def func2(*, opt: float) -> float:
    ...     return opt
    >>> feud.run(
    ...     (func1, func2),
    ...     ["func2", "--opt", "0.12"],
    ...     standalone_mode=False,
    ... )
    0.12

    Running a module of functions, commands or groups.

    >>> import feud
    >>> import types
    >>> feud.run(types)  # doctest: +SKIP

    Running with automatic discovery

    >>> import feud
    >>> def func1(*, opt: int) -> int:
    ...     return opt
    >>> feud.run()  # doctest: +SKIP
    """
    # swap 'obj' and 'args' if empty 'args'
    if (
        args is None
        and isinstance(obj, t.Iterable)
        and not isinstance(obj, dict)
        and all(isinstance(item, str) for item in obj)
    ):
        args = obj
        obj = None

    # use current module if no runner provided
    if obj is None:
        frame = inspect.stack()[1]
        obj = inspect.getmodule(frame[0]) or sys.modules.get("__main__")

    if obj is None:
        msg = (
            "Unable to build command - no runnable object was provided "
            "and no current module can be determined in the present context."
        )
        raise feud.CompilationError(msg)

    # get runner
    runner: click.Command | type[Group] = get_runner(
        obj,
        name=name,
        help=help,
        epilog=epilog,
        config=config,
        warn=warn,
    )

    # set (and unset) feud/user-specified rich-click settings
    with rich_styler(rich_settings or {}):
        return runner(args, **click_kwargs)


def get_runner(
    obj: Runner,
    *,
    name: str | None = None,
    help: str | None = None,  # noqa: A002
    epilog: str | None = None,
    config: Config | None = None,
    convert_func: bool = True,
    warn: bool = True,
) -> click.Command | Group:
    """Generate a :py:class:`click.Command` or :py:class:`.Group` from any
    runnable object.
    """
    provided_kwargs: list[str] = [
        k
        for k, v in {
            "name": name,
            "help": help,
            "epilog": epilog,
            "config": config,
        }.items()
        if v is not None
    ]

    if inspect.isclass(obj) and issubclass(obj, Group):
        if warn and provided_kwargs:
            msg = (
                f"The keyword arguments {provided_kwargs!r} provided to "
                f"feud.run will be ignored.\nConsider redefining the "
                f"specified group with:\n\n\tclass {obj.__name__}(feud.Group"
                f", {', '.join(f'{k}=...' for k in provided_kwargs)})"
                f"\n\nIf this is expected behaviour, this warning can be "
                "silenced by specifying warn=False to feud.run."
            )
            warnings.warn(msg, stacklevel=1)
        return obj
    if isinstance(obj, click.Group):
        if warn and provided_kwargs:
            msg = (
                f"The keyword arguments {provided_kwargs!r} provided to "
                f"feud.run will be ignored.\nConsider providing these "
                f"settings during group construction."
            )
            warnings.warn(msg, stacklevel=1)
        return obj
    if isinstance(obj, click.Command):
        if warn and provided_kwargs:
            msg = (
                f"The keyword arguments {provided_kwargs!r} provided to "
                f"feud.run will be ignored.\nConsider redefining the "
                f"specified command with:\n\n\t@feud.command("
                f"{', '.join(f'{k}=...' for k in provided_kwargs)})"
                f"\n\nIf this is expected behaviour, this warning can be "
                "silenced by specifying warn=False to feud.run."
            )
            warnings.warn(msg, stacklevel=1)
        return obj
    if isinstance(obj, dict):
        items: dict[str, click.Command | type[Group]] = {}
        for k, v in obj.items():
            kwargs: dict[str, t.Any] = {"name": k}
            if isinstance(v, (dict, t.Iterable, types.ModuleType)):
                kwargs["config"] = config
            # set convert_func=False to leave @feud.command to group metaclass
            items[k] = get_runner(v, convert_func=False, warn=False, **kwargs)
        return Group.from_dict(
            items, name=name, help=help, epilog=epilog, config=config
        )
    if isinstance(obj, t.Iterable):
        items: list[click.Command | type[Group]] = []
        for v in obj:
            if isinstance(v, (dict, t.Iterable)):
                msg = (
                    "Groups cannot be constructed from dict or iterable "
                    "objects nested within iterable objects."
                )
                raise feud.exceptions.CompilationError(msg)
            kwargs: dict[str, t.Any] = {}
            if isinstance(v, types.ModuleType):
                kwargs["config"] = config
            # set convert_func=False to leave @feud.command to group metaclass
            runner = get_runner(v, convert_func=False, warn=False, **kwargs)
            items.append(runner)
        return Group.from_iter(
            items, name=name, help=help, epilog=epilog, config=config
        )
    if isinstance(obj, types.ModuleType):
        return Group.from_module(
            obj, name=name, help=help, epilog=epilog, config=config
        )

    if convert_func:
        kwargs: dict[str, t.Any] = {
            "name": name,
            "help": help,
            "epilog": epilog,
            "config": config,
        }
        return command(
            obj,
            **{k: v for k, v in kwargs.items() if v is not None},
        )

    return obj


@contextlib.contextmanager
def rich_styler(
    kwargs: dict[str, t.Any] | None,
    /,
) -> t.Generator[dict[str, t.Any]]:
    """Temporarily applies Feud and user-specified rich-click settings if
    rich-click is installed and rich-click keywords were provided to feud.run.
    """
    # avoid mutating
    kwargs: dict[str, t.Any] = kwargs.copy()

    try:
        if RICH:
            # alias click.rich_click
            rich = click.rich_click

            # check for screaming snake case kwargs that are invalid
            invalid_kwargs: dict[str, t.Any] = {
                k: kwargs.pop(k) for k in kwargs.copy() if not hasattr(rich, k)
            }

            # warn if any invalid kwargs
            if invalid_kwargs:
                msg = (
                    "The following invalid rich-click settings will be "
                    f"ignored: {invalid_kwargs}."
                )
                warnings.warn(msg, stacklevel=1)

            # override: true defaults -> feud defaults -> specified settings
            true_defaults = {k: getattr(rich, k) for k in kwargs}
            kwargs = {**true_defaults, **RICH_DEFAULTS, **kwargs}

            # set rich_click settings
            for k, v in kwargs.items():
                setattr(rich, k, v)

            yield
        else:
            if kwargs:
                msg = (
                    "rich-click settings were provided to feud.run, "
                    "but rich-click is not installed - these settings "
                    "will be ignored."
                )
                warnings.warn(msg, stacklevel=1)

            yield
    finally:
        # restore rich_click defaults
        if RICH:
            for k, v in true_defaults.items():
                setattr(rich, k, v)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import click
import pytest

from feud import typing as t
from feud.config import Config


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        # pydantic extra types
        (t.Color, click.STRING),
        (t.Latitude, click.FLOAT),
        (t.Longitude, click.FLOAT),
        (t.Coordinate, (click.FLOAT, click.FLOAT)),
        (t.CountryAlpha2, click.STRING),
        (t.CountryAlpha3, click.STRING),
        (t.CountryNumericCode, click.STRING),
        (t.CountryShortName, click.STRING),
        (t.CountryOfficialName, click.STRING),
        (t.MacAddress, click.STRING),
        (t.PaymentCardNumber, click.STRING),
        (
            t.PaymentCardBrand,
            lambda x: isinstance(x, click.Choice)
            and x.choices
            == ["American Express", "Mastercard", "Visa", "Mir", "other"],
        ),
        (t.PhoneNumber, click.STRING),
        (t.ABARoutingNumber, click.STRING),
    ],
)
def test_pydantic_extra(
    helpers: type,
    *,
    config: Config,
    annotated: bool,
    hint: t.Any,
    expected: click.ParamType | None,
) -> None:
    helpers.check_get_click_type(
        config=config,
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Officially supported types from the ``pydantic`` package."""

from pydantic import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    AmqpDsn,
    AnyHttpUrl,
    AnyUrl,
    AwareDatetime,
    Base64Bytes,
    Base64Str,
    ByteSize,
    CockroachDsn,
    DirectoryPath,
    EmailStr,
    FilePath,
    FileUrl,
    FiniteFloat,
    FutureDate,
    FutureDatetime,
    HttpUrl,
    ImportString,
    IPvAnyAddress,
    IPvAnyInterface,
    IPvAnyNetwork,
    Json,
    KafkaDsn,
    MariaDBDsn,
    MongoDsn,
    MySQLDsn,
    NaiveDatetime,
    NameEmail,
    NegativeFloat,
    NegativeInt,
    NewPath,
    NonNegativeFloat,
    NonNegativeInt,
    NonPositiveFloat,
    NonPositiveInt,
    PastDate,
    PastDatetime,
    PositiveFloat,
    PositiveInt,
    PostgresDsn,
    RedisDsn,
    SecretBytes,
    SecretStr,
    SkipValidation,
    StrictBool,
    StrictBytes,
    StrictFloat,
    StrictInt,
    StrictStr,
    conbytes,
    condate,
    condecimal,
    confloat,
    confrozenset,
    conint,
    conlist,
    conset,
    constr,
)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import abc
import typing as t

from feud._internal import _command
from feud.config import Config
from feud.core.command import command

try:
    import rich_click as click
except ImportError:
    import click


class GroupBase(abc.ABCMeta):
    def __new__(
        __cls: type[GroupBase],  # noqa: N804
        cls_name: str,
        bases: tuple[type, ...],
        namespace: dict[str, t.Any],
        **kwargs: t.Any,
    ) -> type:  # type[Group], but circular import
        """Metaclass for creating groups.

        Parameters
        ----------
        cls_name:
            The name of the class to be created.

        bases:
            The base classes of the class to be created.

        namespace:
            The attribute dictionary of the class to be created.

        **kwargs:
            Catch-all for any other keyword arguments.

            This can be a combination of:
            - click command/group key-word arguments,
            - feud configuration key-word arguments,
            - a feud configuration object.

        Returns
        -------
        The new class created by the metaclass.
        """
        if bases:
            base_config: Config | None = None
            click_kwargs: dict[str, t.Any] = {}
            subgroups: list[type] = []  # type[Group], but circular import
            commands: list[str] = []

            # extend/inherit information from parent group if subclassed
            help_: str | None = None
            for base in bases:
                if config := getattr(base, "__feud_config__", None):
                    # NOTE: may want **dict(config) depending on behaviour
                    base_config = Config._create(  # noqa: SLF001
                        base=base_config,
                        **config.model_dump(exclude_unset=True),
                    )
                    click_kwargs = {
                        **click_kwargs,
                        **base.__feud_click_kwargs__,
                    }
                    subgroups += [
                        subgroup
                        for subgroup in base.__feud_subgroups__
                        if subgroup not in subgroups
                    ]
                    commands += [
                        cmd
                        for cmd in base.__feud_commands__
                        if cmd not in commands
                    ]
                    help_ = base.__feud_click_kwargs__.get("help")

            # deconstruct base config, override config kwargs and click kwargs
            config_kwargs: dict[str, t.Any] = {}
            for k, v in kwargs.items():
                if k == "config":
                    # NOTE: may want base_config = v depending on behaviour
                    base_config = Config._create(  # noqa: SLF001
                        base=base_config, **v.model_dump(exclude_unset=True)
                    )
                else:
                    d = (
                        config_kwargs
                        if k in Config.model_fields
                        else click_kwargs
                    )
                    d[k] = v

            # sanitize click kwargs
            _command.sanitize_click_kwargs(
                click_kwargs, name=cls_name, help_=help_
            )

            # members to consider as commands
            funcs = {
                name: attr
                for name, attr in namespace.items()
                if callable(attr) and not name.startswith("_")
            }

            # set config and click kwargs
            # (override feud.command decorator settings)
            namespace["__feud_config__"] = Config._create(  # noqa: SLF001
                base=base_config, **config_kwargs
            )
            namespace["__feud_click_kwargs__"] = click_kwargs
            namespace["__feud_subgroups__"] = subgroups
            namespace["__feud_commands__"] = commands + [
                func for func in funcs if func not in commands
            ]

            # auto-generate commands
            for name, func in funcs.items():
                if not isinstance(func, click.Command):
                    namespace[name] = command(
                        func, config=namespace["__feud_config__"]
                    )

        group = super().__new__(__cls, cls_name, bases, namespace)

        if bases:
            # use class-level docstring as help if provided
            if doc := group.__doc__:
                click_kwargs["help"] = doc
            # use __main__ function-level docstring as help if provided
            if doc := group.__main__.__doc__:
                click_kwargs["help"] = doc
            # use class-level click kwargs help if provided
            if doc := kwargs.get("help"):
                click_kwargs["help"] = doc

        return group



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Unit tests for ``feud.decorators``."""

import enum

import pytest

from feud import click
from feud import typing as t
from feud._internal import _decorators


def test_validate_call_single_invalid() -> None:
    """Check output when ``validate_call`` receives a single invalid input
    value.
    """
    name = "func"
    param_renames = {}
    meta_vars = {"arg2": "--arg2"}
    sensitive_vars = {"arg2": False}
    positional = []
    var_positional = None
    pydantic_kwargs = {}

    def f(*, arg2: t.Literal["a", "b", "c"]) -> None:
        pass

    with pytest.raises(click.UsageError) as e:
        _decorators.validate_call(
            f,
            name=name,
            param_renames=param_renames,
            meta_vars=meta_vars,
            sensitive_vars=sensitive_vars,
            positional=positional,
            var_positional=var_positional,
            pydantic_kwargs=pydantic_kwargs,
        )(arg2="invalid")

    assert (
        str(e.value)
        == """
1 validation error for command 'func'
--arg2
  Input should be 'a', 'b' or 'c' [input_value='invalid']
""".strip()
    )


def test_validate_call_multiple_invalid() -> None:
    """Check output when ``validate_call`` receives multiple invalid
    input values.
    """
    name = "func"
    param_renames = {}
    meta_vars = {"0": "ARG1", "arg2": "--arg2"}
    sensitive_vars = {"0": False, "arg2": False}
    positional = []
    var_positional = None
    pydantic_kwargs = {}

    def f(arg1: int, *, arg2: t.Literal["a", "b", "c"]) -> None:
        pass

    with pytest.raises(click.UsageError) as e:
        _decorators.validate_call(
            f,
            name=name,
            param_renames=param_renames,
            meta_vars=meta_vars,
            sensitive_vars=sensitive_vars,
            positional=positional,
            var_positional=var_positional,
            pydantic_kwargs=pydantic_kwargs,
        )("invalid", arg2="invalid")

    assert (
        str(e.value)
        == """
2 validation errors for command 'func'
ARG1
  Input should be a valid integer, unable to parse string as an integer [input_value='invalid']
--arg2
  Input should be 'a', 'b' or 'c' [input_value='invalid']
""".strip()  # noqa: E501
    )


def test_validate_call_list() -> None:
    """Check output when ``validate_call`` receives an invalid input value for
    a list argument.
    """
    name = "func"
    param_renames = {}
    meta_vars = {"0": "[ARG1]..."}
    sensitive_vars = {"0": False}
    positional = []
    var_positional = None
    pydantic_kwargs = {}

    def f(arg1: list[t.conint(multiple_of=2)]) -> None:
        pass

    with pytest.raises(click.UsageError) as e:
        _decorators.validate_call(
            f,
            name=name,
            param_renames=param_renames,
            meta_vars=meta_vars,
            sensitive_vars=sensitive_vars,
            positional=positional,
            var_positional=var_positional,
            pydantic_kwargs=pydantic_kwargs,
        )([1, 2, 3])

    assert (
        str(e.value)
        == """
2 validation errors for command 'func'
[ARG1]... [0]
  Input should be a multiple of 2 [input_value=1]
[ARG1]... [2]
  Input should be a multiple of 2 [input_value=3]
""".strip()
    )


def test_validate_call_enum() -> None:
    """Check output when ``validate_call`` receives an invalid input value
    for an enum parameter.
    """
    name = "func"
    param_renames = {}
    meta_vars = {"arg2": "--arg2"}
    sensitive_vars = {"arg2": False}
    positional = []
    var_positional = None
    pydantic_kwargs = {}

    class Choice(enum.Enum):
        A = "a"
        B = "b"
        C = "c"

    def f(*, arg2: Choice) -> None:
        pass

    with pytest.raises(click.UsageError) as e:
        _decorators.validate_call(
            f,
            name=name,
            param_renames=param_renames,
            meta_vars=meta_vars,
            sensitive_vars=sensitive_vars,
            positional=positional,
            var_positional=var_positional,
            pydantic_kwargs=pydantic_kwargs,
        )(arg2="invalid")

    assert (
        str(e.value)
        == """
1 validation error for command 'func'
--arg2
  Input should be 'a', 'b' or 'c' [input_value='invalid']
""".strip()
    )


def test_validate_call_datetime() -> None:
    """Check output when ``validate_call`` receives an invalid input value
    for a datetime parameter.
    """
    name = "func"
    param_renames = {}
    meta_vars = {"time": "--time"}
    sensitive_vars = {"time": False}
    positional = []
    var_positional = None
    pydantic_kwargs = {}

    def f(*, time: t.FutureDatetime) -> None:
        pass

    with pytest.raises(click.UsageError) as e:
        _decorators.validate_call(
            f,
            name=name,
            param_renames=param_renames,
            meta_vars=meta_vars,
            sensitive_vars=sensitive_vars,
            positional=positional,
            var_positional=var_positional,
            pydantic_kwargs=pydantic_kwargs,
        )(time=t.datetime.now())

    assert str(e.value).startswith(
        """
1 validation error for command 'func'
--time
  Input should be in the future
""".strip()
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import typing as t

import pytest


class Helpers:
    @staticmethod
    def annotate(hint: t.Any) -> t.Annotated[t.Any, "annotation"]:
        return t.Annotated[hint, "annotation"]


@pytest.fixture(scope="module")
def helpers() -> type[Helpers]:
    return Helpers



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Tasks for bumping the package version."""

import os
import re
from pathlib import Path

from invoke.config import Config
from invoke.tasks import task


@task
def build(c: Config, *, v: str) -> None:
    """Build release."""
    root: Path = Path(os.getcwd())

    # bump Sphinx documentation version - docs/source/conf.py
    conf_path: Path = root / "docs" / "source" / "conf.py"
    with open(conf_path) as f:
        conf: str = f.read()
    with open(conf_path, "w") as f:
        f.write(re.sub(r'release = ".*"', f'release = "{v}"', conf))

    # bump package version - feud/version.py)
    init_path: Path = root / "feud" / "version.py"
    with open(init_path) as f:
        init: str = f.read()
    with open(init_path, "w") as f:
        f.write(re.sub(r'VERSION = ".*"', f'VERSION = "{v}"', init))

    # bump project version - pyproject.toml
    c.run(f"poetry version -q {v}")



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import typing as t

import pytest

import feud


def test_union(capsys: pytest.CaptureFixture) -> None:
    @feud.command
    def f(
        *,
        opt1: int | float,
        opt2: t.Union[int, float],
        opt3: str | int | None,
        opt4: t.Optional[t.Union[str, int]],
        opt5: t.Union[int, t.Union[float, str]],
        opt6: int | None,
        opt7: str | t.Annotated[str, "annotated"],
    ) -> None:
        pass

    with pytest.raises(SystemExit):
        f(["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS]

Options:
  --opt1 INTEGER | FLOAT         [required]
  --opt2 INTEGER | FLOAT         [required]
  --opt3 TEXT | INTEGER          [required]
  --opt4 TEXT | INTEGER          [required]
  --opt5 INTEGER | FLOAT | TEXT  [required]
  --opt6 INTEGER                 [required]
  --opt7 TEXT                    [required]
  --help                         Show this message and exit.
    """.strip()
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Officially supported types.

Note that other arbitrary types may work with Feud,
but those imported within this module are the officially supported types.
"""

from feud.typing.custom import *
from feud.typing.pydantic import *
from feud.typing.pydantic_extra_types import *
from feud.typing.stdlib import *
from feud.typing.typing import *



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import click
import pytest

from feud import typing as t
from feud._internal._types.click import Union
from feud.config import Config


def annotate(hint: t.Any) -> t.Annotated[t.Any, "annotation"]:
    return t.Annotated[hint, "annotation"]


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (t.Any, None),
        (t.Text, click.STRING),
        (t.Pattern, None),
        (t.Optional[int], click.INT),
        (t.Optional[annotate(int)], click.INT),
        (
            t.Literal["a", "b"],
            lambda x: isinstance(x, click.Choice) and x.choices == ["a", "b"],
        ),
        (t.Tuple, None),
        (t.Tuple[int], (click.INT,)),
        (t.Tuple[annotate(int)], (click.INT,)),
        (t.Tuple[int, str], (click.INT, click.STRING)),
        (t.Tuple[annotate(int), annotate(str)], (click.INT, click.STRING)),
        (t.Tuple[int, ...], click.INT),
        (t.Tuple[annotate(int), ...], click.INT),
        (t.List, None),
        (t.List[int], click.INT),
        (t.List[annotate(int)], click.INT),
        (t.Set, None),
        (t.Set[int], click.INT),
        (t.Set[annotate(int)], click.INT),
        (t.FrozenSet, None),
        (t.FrozenSet[int], click.INT),
        (t.FrozenSet[annotate(int)], click.INT),
        (t.Deque, None),
        (t.Deque[int], click.INT),
        (t.Deque[annotate(int)], click.INT),
        (t.NamedTuple("Point", x=int, y=str), (click.INT, click.STRING)),
        (
            t.NamedTuple("Point", x=annotate(int), y=annotate(str)),
            (click.INT, click.STRING),
        ),
        (
            t.Union[int, str],
            lambda x: isinstance(x, Union)
            and x.types == [click.INT, click.STRING],
        ),
    ],
)
def test_typing(
    helpers: type,
    *,
    config: Config,
    annotated: bool,
    hint: t.Any,
    expected: click.ParamType | None,
) -> None:
    helpers.check_get_click_type(
        config=config,
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import dataclasses
import enum
import inspect
import typing as t

try:
    import rich_click as click
except ImportError:
    import click

from feud._internal import _decorators, _inflect, _types
from feud.config import Config

CONTEXT_PARAM = "ctx"


class ParameterType(enum.Enum):
    OPTION = enum.auto()
    ARGUMENT = enum.auto()


@dataclasses.dataclass
class ParameterSpec:
    type: ParameterType | None = None  # noqa: A003
    hint: type | None = None
    args: t.Iterable[t.Any] = dataclasses.field(default_factory=list)
    kwargs: dict[str, t.Any] = dataclasses.field(default_factory=dict)


class NameDict(t.TypedDict):
    command: str | None
    params: dict[str, str]


@dataclasses.dataclass
class CommandState:
    config: Config
    click_kwargs: dict[str, t.Any]
    is_group: bool
    names: dict[str, NameDict]  # key: parameter name
    aliases: dict[str, str | list[str]]  # key: parameter name
    envs: dict[str, str]  # key: parameter name
    overrides: dict[str, click.Parameter]  # key: parameter name
    pass_context: bool = False
    # below keys are parameter name
    arguments: dict[str, ParameterSpec] = dataclasses.field(
        default_factory=dict
    )
    options: dict[str, ParameterSpec] = dataclasses.field(default_factory=dict)

    def decorate(self: CommandState, func: t.Callable) -> click.Command:
        meta_vars: dict[str, str] = {}
        sensitive_vars: dict[str, bool] = {}
        positional: list[str] = []
        var_positional: str | None = None
        params: list[click.Parameter] = []

        sig: inspect.signature = inspect.signature(func)

        for i, (param_name, param_spec) in enumerate(sig.parameters.items()):
            # store names of positional arguments
            if param_spec.kind in (
                param_spec.POSITIONAL_ONLY,
                param_spec.POSITIONAL_OR_KEYWORD,
            ):
                positional.append(param_name)
            # store name of variable positional arguments (i.e. *args)
            if param_spec.kind == param_spec.VAR_POSITIONAL:
                var_positional = param_name
            # ignore variable keyword arguments (i.e. **kwargs)
            if param_spec.kind == param_spec.VAR_KEYWORD:
                continue

            # create Click parameters
            sensitive: bool = False
            if self.pass_context and i == 0:
                continue
            if param_name in self.overrides:
                param: click.Parameter = self.overrides[param_name]
                sensitive = param.hide_input or param.envvar
            elif param_name in self.arguments:
                spec = self.arguments[param_name]
                spec.kwargs["type"] = _types.click.get_click_type(
                    spec.hint, config=self.config
                )
                param = click.Argument(spec.args, **spec.kwargs)
            elif param_name in self.options:
                spec = self.options[param_name]
                spec.kwargs["type"] = _types.click.get_click_type(
                    spec.hint, config=self.config
                )
                param = click.Option(spec.args, **spec.kwargs)
                hide_input = spec.kwargs.get("hide_input")
                envvar = spec.kwargs.get("envvar")
                sensitive = hide_input or envvar

            # get renamed parameter if @feud.rename used
            name: str = self.names["params"].get(param_name, param_name)

            # set parameter name
            param.name = name

            # get meta vars and identify sensitive parameters for validate_call
            meta_vars[name] = self.get_meta_var(param)
            sensitive_vars[name] = sensitive

            # add the parameter
            params.append(param)

        # add any overrides that don't appear in function signature
        # e.g. version_option or anything else
        for param_name, param in self.overrides.items():
            if param_name not in sig.parameters:
                params.append(param)

        # rename command if @feud.rename used
        if command_rename := self.names["command"]:
            self.click_kwargs = {**self.click_kwargs, "name": command_rename}

        command = _decorators.validate_call(
            func,
            name=self.click_kwargs["name"],
            param_renames=self.names["params"],
            meta_vars=meta_vars,
            sensitive_vars=sensitive_vars,
            positional=positional,
            var_positional=var_positional,
            pydantic_kwargs=self.config.pydantic_kwargs,
        )

        if self.pass_context:
            command = click.pass_context(command)

        constructor = click.group if self.is_group else click.command
        command = constructor(**self.click_kwargs)(command)

        command.params = params

        return command

    def get_meta_var(self: CommandState, param: click.Parameter) -> str:
        match param:
            case click.Argument():
                return param.make_metavar()
            case click.Option():
                return param.opts[0]


def pass_context(sig: inspect.Signature) -> bool:
    """Determine whether or not ``click.pass_context`` should be called.

    Context is passed ff the first parameter if the function is named ``ctx``.
    """
    param_name: str | None = dict(enumerate(sig.parameters)).get(0)
    return param_name == CONTEXT_PARAM


def get_option(name: str, *, hint: type, negate_flags: bool) -> str:
    """Convert a name into a command-line option.

    Additionally negates the option if a boolean flag is provided
    and ``negate_flags`` is ``True``, and returns a joint declaration.

    Example
    -------
    >>> get_option("opt-name", hint=bool, negate_flags=True)
    "--opt-name/--no-opt-name"
    """
    option: str = _inflect.optionize(name)
    base_type, _, _, _ = _types.click.get_base_type(hint)
    if base_type is bool and negate_flags:
        negated_option: str = _inflect.negate_option(option)
        return f"{option}/{negated_option}"
    return option


def get_alias(alias: str, *, hint: type, negate_flags: bool) -> str:
    """Negate an alias for a boolean flag and returns a joint declaration
    if ``negate_flags`` is ``True``.

    Example
    -------
    >>> get_alias("-a", hint=bool, negate_flags=True)
    "-a/--no-a"
    >>> get_alias("-a", hint=bool, negate_flags=False)
    "-a"
    >>> get_alias("-a", hint=str, negate_flags=True)
    "-a"
    """
    base_type, _, _, _ = _types.click.get_base_type(hint)
    if base_type is bool and negate_flags:
        negated_alias: str = _inflect.negate_alias(alias)
        return f"{alias}/{negated_alias}"
    return alias


def sanitize_click_kwargs(
    click_kwargs: dict[str, t.Any], *, name: str, help_: str | None = None
) -> None:
    """Sanitize click command/group arguments.

    Removes the ``commands`` argument and sets a name if not present.
    """
    # remove pre-specified commands
    click_kwargs.pop("commands", None)
    # sanitize the provided name
    # (only necessary for auto-naming a Group by class name)
    click_kwargs["name"] = click_kwargs.get("name", _inflect.sanitize(name))
    # set help if provided
    if help_:
        click_kwargs["help"] = help_



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""This is a module."""

import types

import feud
from feud import click


def func1(*, opt: bool) -> bool:
    """This is the first function.\f

    Parameters
    ----------
    opt:
        This is an option.
    """
    return opt


@feud.command
def command(*, opt: bool) -> bool:
    """This is a command.\f

    Parameters
    ----------
    opt:
        This is an option.
    """
    return opt


class Group(feud.Group, name="feud-group"):
    """This is a Feud group."""

    def func(*, opt: bool) -> bool:
        """This is a function in the group.\f

        Parameters
        ----------
        opt:
            This is an option.
        """
        return opt


class Subgroup(feud.Group, name="feud-subgroup"):
    """This is a subgroup."""

    def func(*, opt: bool) -> bool:
        """This is a function in the subgroup.\f

        Parameters
        ----------
        opt:
            This is an option.
        """
        return opt


Group.register(Subgroup)

click_group: click.Group = feud.compile(
    types.new_class(
        "ClickGroup",
        bases=(Group,),
        kwds={
            "name": "click-group",
            "help": "This is a Click group.",
        },
    )
)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import collections
import datetime
import decimal
import enum
import functools as ft
import inspect
import pathlib
import types
import typing as t
import uuid

import click
import pydantic as pyd
from annotated_types import Interval

from feud.config import Config

PATH_TYPES = (
    pathlib.Path,
    pathlib.PosixPath,
    pathlib.PurePath,
    pathlib.PurePosixPath,
    pathlib.PureWindowsPath,
    pathlib.WindowsPath,
)

DATE_TYPES = (
    datetime.date,
    pyd.PastDate,
    pyd.FutureDate,
)

TIME_TYPES = (datetime.time,)

DATETIME_TYPES = (
    datetime.datetime,
    pyd.PastDatetime,
    pyd.FutureDatetime,
    pyd.AwareDatetime,
    pyd.NaiveDatetime,
)

TIMEDELTA_TYPES = (datetime.timedelta,)

BASE_TYPES = {
    str: click.STRING,
    int: click.INT,
    float: click.FLOAT,
    decimal.Decimal: click.FLOAT,
    bool: click.BOOL,
    uuid.UUID: click.UUID,
    **{path_type: click.Path() for path_type in PATH_TYPES},
}

try:
    from pydantic_extra_types.color import Color
    from pydantic_extra_types.coordinate import (
        Coordinate,
        Latitude,
        Longitude,
    )
    from pydantic_extra_types.country import (
        CountryAlpha2,
        CountryAlpha3,
        CountryNumericCode,
        CountryOfficialName,
        CountryShortName,
    )
    from pydantic_extra_types.mac_address import MacAddress
    from pydantic_extra_types.payment import (
        PaymentCardNumber,
    )
    from pydantic_extra_types.phone_numbers import PhoneNumber
    from pydantic_extra_types.routing_number import ABARoutingNumber

    # NOTE: PaymentCardBrand is skipped as it is just an enum
    EXTRA_TYPES = {
        Color: click.STRING,
        Coordinate: (click.FLOAT, click.FLOAT),
        Latitude: click.FLOAT,
        Longitude: click.FLOAT,
        CountryAlpha2: click.STRING,
        CountryAlpha3: click.STRING,
        CountryNumericCode: click.STRING,
        CountryOfficialName: click.STRING,
        CountryShortName: click.STRING,
        MacAddress: click.STRING,
        PaymentCardNumber: click.STRING,
        PhoneNumber: click.STRING,
        ABARoutingNumber: click.STRING,
    }
except ImportError:
    EXTRA_TYPES = {}

COLLECTION_TYPES = (
    tuple,
    list,
    set,
    frozenset,
    collections.deque,
)

DEFAULT_TYPE = None

PathType = t.Union[*PATH_TYPES]
ClickType = t.Union[*BASE_TYPES.values(), type(DEFAULT_TYPE)]
AnnotatedArgDict = t.Dict[int, t.Any]


class DateTime(click.DateTime):
    def __init__(
        self: DateTime,
        *args: t.Any,
        datetime_type: type,
        show_default_format: bool,
        **kwargs: t.Any,
    ) -> DateTime:
        self._datetime_type = datetime_type
        self._show_default_format = show_default_format
        self.name = datetime_type.__name__
        super().__init__(*args, **kwargs)

    def get_metavar(
        self: DateTime,
        param: click.Parameter,  # noqa: ARG002
    ) -> str:
        return (
            f"[{'|'.join(self.formats)}]"
            if self._show_default_format
            else self.name.upper()
        )

    def _try_to_convert_date(
        self: DateTime,
        value: t.Any,
        format: str,  # noqa: A002, ARG002
    ) -> t.Any | None:
        try:
            return pyd.TypeAdapter(self._datetime_type).validate_python(value)
        except pyd.ValidationError:
            return None


class Union(click.ParamType):
    def __init__(
        self: DateTime,
        *args: t.Any,
        types: list[click.ParamType],
        **kwargs: t.Any,
    ) -> DateTime:
        self.types = types
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_metavar(
        click_type: click.ParamType | None,
        param: click.Parameter,
    ) -> str:
        if click_type:
            return click_type.get_metavar(param) or click_type.name.upper()
        return None

    def get_metavar(self: DateTime, param: click.Parameter) -> str:
        metavars = [
            metavar
            for click_type in self.types
            if (metavar := self._get_metavar(click_type, param))
        ]
        unique_metavars = list(dict.fromkeys(metavars))
        return " | ".join(unique_metavars)


def get_click_type(hint: t.Any, *, config: Config) -> ClickType | None:
    base_type, base_args, _, _ = get_base_type(hint)
    origin_type = t.get_origin(base_type)

    click_type = resolve_collection(origin_type, args=base_args, config=config)
    if click_type and origin_type in COLLECTION_TYPES:
        return click_type

    click_type = resolve_collection(base_type, args={}, config=config)
    if click_type and is_namedtuple(base_type):
        return click_type

    return resolve_type(hint, config=config)


def resolve_type(hint: t.Any, *, config: Config) -> ClickType:
    base_type, base_args, parent_type, parent_args = get_base_type(hint)
    if t.get_origin(parent_type) is t.Annotated:
        click_type = resolve_annotated(base_type, parent_args=parent_args)
        if click_type:
            return click_type
    if t.get_origin(base_type) is t.Literal:
        return click.Choice(list(map(str, base_args.values())))
    if t.get_origin(base_type) in (t.Union, types.UnionType):
        # only try to determine type for
        # t.Optional[t.Any] / t.Union[t.Any, None]
        arg_values = base_args.values()
        if len(base_args) == 2 and type(None) in arg_values:
            non_none = next(arg for arg in arg_values if arg is not type(None))
            return get_click_type(non_none, config=config)
        # t.Union with more than one non-None argument
        base_types = list(
            map(
                ft.partial(get_click_type, config=config),
                base_args.values(),
            )
        )
        return Union(types=base_types)
    if inspect.isclass(base_type):
        if issubclass(base_type, enum.Enum):
            return click.Choice([str(e.value) for e in base_type])
        if base_type in DATE_TYPES:
            return DateTime(
                formats=["YYYY-MM-DD"],
                datetime_type=datetime.date,
                show_default_format=config.show_help_datetime_formats,
            )
        if base_type in TIME_TYPES:
            return DateTime(
                formats=["HH:MM[:SS[.ffffff]][Z or [±]HH[:]MM]"],
                datetime_type=datetime.time,
                show_default_format=config.show_help_datetime_formats,
            )
        if base_type in DATETIME_TYPES:
            return DateTime(
                formats=["YYYY-MM-DD[T]HH:MM[:SS[.ffffff]][Z or [±]HH[:]MM]"],
                datetime_type=datetime.datetime,
                show_default_format=config.show_help_datetime_formats,
            )
        if base_type in TIMEDELTA_TYPES:
            return DateTime(
                formats=[
                    "[-][DD ][HH:MM]SS[.ffffff]",
                    "[±]P[DD]DT[HH]H[MM]M[SS]S",
                ],
                datetime_type=datetime.timedelta,
                show_default_format=config.show_help_datetime_formats,
            )
        if base_type in EXTRA_TYPES:
            return EXTRA_TYPES[base_type]
    return BASE_TYPES.get(base_type, DEFAULT_TYPE)


def get_arg_dict(hint: t.Annotated[t.Any, ...]) -> AnnotatedArgDict:
    """Convert the arguments of an annotated type into a dictionary.

    Dictionary is keyed by argument index.
    """
    return dict(enumerate(t.get_args(hint)))


def get_base_type(
    hint: t.Any,
    *,
    parent_args: AnnotatedArgDict | None = None,
    parent_type: t.Any | None = None,
) -> tuple[t.Any, AnnotatedArgDict, t.Any, AnnotatedArgDict]:
    """Retrieve the inner type and arguments of a type.

    Can be annotated or non-annotated. Also returns outer type and arguments.

    Examples
    --------
    >>> import typing as t
    >>> get_base_type(t.Annotated[t.Tuple[int, ...], "annotation"])
    (
        typing.Tuple[int, ...],
        {0: <class 'int'>, 1: Ellipsis},
        typing.Annotated[typing.Tuple[int, ...], 'annotation'],
        {0: typing.Tuple[int, ...], 1: 'annotation'}
    )
    """
    args = get_arg_dict(hint)
    if t.get_origin(hint) is t.Annotated:
        return get_base_type(args[0], parent_type=hint, parent_args=args)
    return hint, args, parent_type, parent_args


def is_collection_type(hint: t.Any) -> tuple[bool, t.Any | None]:
    """Check for collection types and returns their base type.

    Below is a list of valid collection types.
    Note that fixed size tuples (and typing.NamedTuple) are
    excluded as they have a special meaning in Click.

    - ``tuple``
    - ``typing.Tuple``
    - ``typing.Tuple[typing.Any, ...]``
    - ``list``
    - ``typing.List``
    - ``typing.List[typing.Any]``
    - ``set``
    - ``typing.Set``
    - ``typing.Set[typing.Any]``
    - ``frozenset``
    - ``typing.FrozenSet``
    - ``typing.FrozenSet[t.Any]``
    - ``collections.deque``
    - ``typing.Deque``
    - ``typing.Deque[t.Any]``

    Parameters
    ----------
    hint:
        Any type.

    Examples
    --------
    >>> import feud.typing as t
    >>> from feud._internal import _types
    >>> _types.click.is_collection_type(list)
    (True, None)
    >>> _types.click.is_collection_type(t.Tuple[int, ...])
    (True, int)
    >>> _types.click.is_collection_type(t.Tuple[int, str])
    (False, None)
    >>> _types.click.is_collection_type(str)
    (False, None)
    """
    base_type, base_args, _, _ = get_base_type(hint)

    if base_type in COLLECTION_TYPES:
        return True, None

    origin = t.get_origin(base_type)
    if origin in COLLECTION_TYPES:
        if origin is tuple:
            if len(base_args) == 0:
                # typing.Tuple
                return True, None
            if len(base_args) == 2 and base_args.get(1) is Ellipsis:
                # typing.Tuple[typing.Any, ...]
                return True, base_args.get(0)
            return False, None
        return True, base_args.get(0)

    return False, None


def resolve_collection(
    hint: t.Any, *, args: AnnotatedArgDict, config: Config
) -> ClickType | None:
    """Resolve a Click type for the provided collection type.

    See `is_collection_type` for more information about collection types.

    Examples
    --------
    >>> import feud.typing as t
    >>> from feud.config import Config
    >>> from feud._internal import _types
    >>> _types.click.resolve_collection(
    ...     tuple,
    ...     args={0: t.conint(ge=0, le=3), 1: ...},
    ...     config=Config._create()
    ... )
    click.IntRange(min=0, max=3, min_open=False, max_open=False)
    """
    if hint is tuple and len(args):
        # arbitrary size tuple - t.Tuple[t.Any, ...]
        if args.get(1) is Ellipsis:
            return resolve_type(args[0], config=config)
        # fixed size tuple, e.g. t.Tuple[t.Any, t.Any]
        return tuple(
            map(ft.partial(resolve_type, config=config), args.values())
        )
    if is_namedtuple(hint):
        return tuple(
            map(
                ft.partial(resolve_type, config=config),
                hint.__annotations__.values(),
            )
        )
    if hint in (list, set, frozenset, collections.deque):
        # Type[t.Any]
        return resolve_type(args.get(0), config=config)
    return None


def resolve_annotated(
    base_type: t.Any, *, parent_args: AnnotatedArgDict
) -> ClickType | None:
    arg_list = list(parent_args.values())  # noqa: F841
    two_field_subtype = t.Annotated[*arg_list[:2]]
    # integer types
    if two_field_subtype == pyd.PositiveInt:
        return click.IntRange(min=0, min_open=True)
    if two_field_subtype == pyd.NonNegativeInt:
        return click.IntRange(min=0, min_open=False)
    if two_field_subtype == pyd.NegativeInt:
        return click.IntRange(max=0, max_open=True)
    if two_field_subtype == pyd.NonPositiveInt:
        return click.IntRange(max=0, max_open=False)
    # float types
    if two_field_subtype == pyd.PositiveFloat:
        return click.FloatRange(min=0, min_open=True)
    if two_field_subtype == pyd.NonNegativeFloat:
        return click.FloatRange(min=0, min_open=False)
    if two_field_subtype == pyd.NegativeFloat:
        return click.FloatRange(max=0, max_open=True)
    if two_field_subtype == pyd.NonPositiveFloat:
        return click.FloatRange(max=0, max_open=False)
    # int / float range types
    if is_pyd_conint(base_type, parent_args):
        return get_click_range_type(parent_args, range_type=click.IntRange)
    if is_pyd_confloat(base_type, parent_args):
        return get_click_range_type(parent_args, range_type=click.FloatRange)
    if is_pyd_condecimal(base_type, parent_args):
        return get_click_range_type(parent_args, range_type=click.FloatRange)
    # file / directory types
    if two_field_subtype == pyd.FilePath:
        return click.Path(exists=True, dir_okay=False)
    if two_field_subtype == pyd.DirectoryPath:
        return click.Path(exists=True, file_okay=False)
    if base_type in PATH_TYPES:
        return click.Path()
    return None


def is_pyd_conint(base_type: t.Any, parent_args: AnnotatedArgDict) -> bool:
    return base_type is int and isinstance(parent_args.get(2), Interval)


def is_pyd_confloat(base_type: t.Any, parent_args: AnnotatedArgDict) -> bool:
    return base_type is float and isinstance(parent_args.get(2), Interval)


def is_pyd_condecimal(base_type: t.Any, parent_args: AnnotatedArgDict) -> bool:
    return base_type is decimal.Decimal and isinstance(
        parent_args.get(2), Interval
    )


def is_namedtuple(hint: t.Any) -> bool:
    if hint is None:
        return False
    if not inspect.isclass(hint):
        return False
    return issubclass(hint, tuple) and hasattr(hint, "_fields")


def get_click_range_type(
    args: tuple, *, range_type: type[click.IntRange] | type[click.FloatRange]
) -> click.IntRange | click.FloatRange:
    min_, max_ = None, None
    min_open, max_open = False, False
    interval: Interval = args.get(2)
    if interval.gt is not None:
        min_ = interval.gt
        min_open = True
    if interval.ge is not None:
        min_ = interval.ge
        min_open = False
    if interval.lt is not None:
        max_ = interval.lt
        max_open = True
    if interval.le is not None:
        max_ = interval.le
        max_open = False
    return range_type(min=min_, max=max_, min_open=min_open, max_open=max_open)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import datetime
import enum
import typing as t


def convert_default(default: t.Any) -> t.Any:
    if isinstance(default, enum.Enum):
        return convert_default(default.value)
    if isinstance(default, datetime.datetime):
        # this would be caught by isinstance(default, datetime.date) otherwise
        return default
    if isinstance(default, (datetime.date, datetime.time)):
        return str(default)
    if isinstance(default, (set, frozenset, tuple, list)):
        return type(default)(map(convert_default, default))
    return default



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import feud
from feud import click


def test_metaclass_default() -> None:
    class Test(feud.Group):
        def command(*, opt: bool) -> None:
            pass

    assert Test.__feud_config__ == feud.config()
    assert Test.__feud_click_kwargs__ == {"name": "test"}

    command = Test.command
    assert isinstance(command, click.Command)

    arg = command.params[0]
    assert arg.opts == ["--opt"]
    assert arg.secondary_opts == ["--no-opt"]


def test_metaclass_base_config() -> None:
    config = feud.config(negate_flags=False)

    class Test(feud.Group, config=config):
        def command(*, opt: bool) -> None:
            pass

    assert Test.__feud_config__ == config
    assert Test.__feud_click_kwargs__ == {"name": "test"}

    command = Test.command
    assert isinstance(command, click.Command)

    arg = command.params[0]
    assert arg.opts == ["--opt"]
    assert arg.secondary_opts == []


def test_metaclass_feud_kwargs() -> None:
    class Test(feud.Group, negate_flags=False):
        def command(*, opt: bool) -> None:
            pass

    assert Test.__feud_config__ == feud.config(negate_flags=False)
    assert Test.__feud_click_kwargs__ == {"name": "test"}

    command = Test.command
    assert isinstance(command, click.Command)

    arg = command.params[0]
    assert arg.opts == ["--opt"]
    assert arg.secondary_opts == []


def test_metaclass_config_with_feud_kwargs() -> None:
    config = feud.config(negate_flags=False)

    class Test(feud.Group, config=config, negate_flags=True):
        def command(*, opt: bool) -> None:
            pass

    assert Test.__feud_config__ == feud.config(negate_flags=True)
    assert Test.__feud_click_kwargs__ == {"name": "test"}

    command = Test.command
    assert isinstance(command, click.Command)

    arg = command.params[0]
    assert arg.opts == ["--opt"]
    assert arg.secondary_opts == ["--no-opt"]


def test_metaclass_click_kwargs() -> None:
    name = "custom"
    epilog = "Visit https://www.com for more information."

    class Test(feud.Group, name=name, epilog=epilog):
        def command(*, opt: bool) -> None:
            pass

    assert Test.__feud_config__ == feud.config()
    assert Test.__feud_click_kwargs__ == {"name": name, "epilog": epilog}

    group = feud.compile(Test)
    assert group.name == name
    assert group.epilog == epilog

    # group-level click kwargs should not be propagated to commands
    command = Test.command
    assert isinstance(command, click.Command)
    assert command.name == "command"
    assert command.epilog is None

    arg = command.params[0]
    assert arg.opts == ["--opt"]
    assert arg.secondary_opts == ["--no-opt"]


def test_metaclass_config_with_feud_kwargs_and_click_kwargs() -> None:
    name = "custom"
    epilog = "Visit https://www.com for more information."
    config = feud.config(negate_flags=False)

    class Test(
        feud.Group, name=name, epilog=epilog, config=config, negate_flags=True
    ):
        def command(*, opt: bool) -> None:
            pass

    assert Test.__feud_config__ == feud.config(negate_flags=True)
    assert Test.__feud_click_kwargs__ == {"name": name, "epilog": epilog}

    group = feud.compile(Test)
    assert group.name == name
    assert group.epilog == epilog

    # group-level click kwargs should not be propagated to commands
    command = Test.command
    assert isinstance(command, click.Command)
    assert command.name == "command"
    assert command.epilog is None

    arg = command.params[0]
    assert arg.opts == ["--opt"]
    assert arg.secondary_opts == ["--no-opt"]



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import re
import unicodedata

__all__ = ["sanitize", "optionize", "negate_option", "negate_alias"]


def transliterate(string: str) -> str:
    """Replace non-ASCII characters with an ASCII approximation. If no
    approximation exists, the non-ASCII character is ignored. The string must
    be ``unicode``.

    Example
    -------
    >>> transliterate("älämölö")
    "alamolo"
    >>> transliterate("Ærøskøbing")
    "rskbing"

    Source code from inflection package (https://github.com/jpvanhal/inflection).

        Copyright (C) 2012-2020 Janne Vanhala

        Permission is hereby granted, free of charge, to any person obtaining
        a copy of this software and associated documentation files
        (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
        NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
        BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
        ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
        CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
    """
    normalized = unicodedata.normalize("NFKD", string)
    return normalized.encode("ascii", "ignore").decode("ascii")


def parameterize(string: str, separator: str = "-") -> str:
    """Replace special characters in a string so that it may be used as part
    of a 'pretty' URL.

    Example
    -------
    >>> parameterize(u"Donald E. Knuth")
    "donald-e-knuth"

    Source code from inflection package (https://github.com/jpvanhal/inflection).

        Copyright (C) 2012-2020 Janne Vanhala

        Permission is hereby granted, free of charge, to any person obtaining
        a copy of this software and associated documentation files
        (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
        NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
        BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
        ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
        CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
    """
    string = transliterate(string)
    # Turn unwanted chars into the separator
    string = re.sub(r"(?i)[^a-z0-9\-_]+", separator, string)
    if separator:
        re_sep = re.escape(separator)
        # No more than one of the separator in a row.
        string = re.sub(r"%s{2,}" % re_sep, separator, string)
        # Remove leading/trailing separator.
        string = re.sub(rf"(?i)^{re_sep}|{re_sep}$", "", string)

    return string.lower()


def sanitize(name: str) -> str:
    """Sanitizes a string for preparation of usage as a command-line option.

    Example
    -------
    >>> sanitize("---a@b_c--")
    "a-b_c"
    """
    name = parameterize(name)
    return re.sub(r"^-*(.*)", r"\1", name)


def optionize(name: str) -> str:
    """Sanitizes a string and converts it into a command-line option.

    Example
    -------
    >>> optionize("opt_name")
    "--opt-name"
    """
    return "--" + sanitize(name)


def negate_option(option: str) -> str:
    """Negates a command-line option (for boolean flags).

    Example
    -------
    >>> negate_option("--opt")
    "--no-opt"
    """
    return "--no" + option.removeprefix("-")


def negate_alias(alias: str) -> str:
    """Negates an alias for a boolean flag.

    Example
    -------
    >>> negate_alias("-a")
    "--no-a"
    """
    return "--no" + alias



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Tasks for running tests."""

from __future__ import annotations

from invoke.config import Config
from invoke.tasks import task


@task
def install(c: Config) -> None:
    """Install package with core and test dependencies."""
    c.run(
        "poetry install --sync --only base,main,tests -E extra-types -E email"
    )


@task
def doctest(c: Config) -> None:
    """Run doctests."""
    # skip:
    # - feud/click/context.py
    # - feud/decorators.py
    files: list[str] = [
        "feud/config.py",
        "feud/core/__init__.py",
        "feud/core/command.py",
        "feud/core/group.py",
    ]
    c.run(f"poetry run python -m doctest {' '.join(files)}")


@task
def unit(c: Config, *, cov: bool = False) -> None:
    """Run unit tests."""
    command: str = "poetry run pytest tests/"

    if cov:
        command = f"{command} --cov feud --cov-report xml"

    c.run(command)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import click
import pytest

from feud import typing as t
from feud.config import Config


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        # counter types
        (t.Counter, click.INT),
        (
            t.concounter(ge=0, le=3),
            lambda x: isinstance(x, click.IntRange)
            and x.min == 0
            and x.min_open is False
            and x.max == 3
            and x.max_open is False,
        ),
    ],
)
def test_custom(
    helpers: type,
    *,
    config: Config,
    annotated: bool,
    hint: t.Any,
    expected: click.ParamType | None,
) -> None:
    helpers.check_get_click_type(
        config=config,
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Supplementary decorators for modifying commands or their parameters."""

from __future__ import annotations

import inspect
import re
import typing as t

import pydantic as pyd

from feud._internal import _command
from feud.exceptions import CompilationError

__all__ = ["alias", "env", "rename"]


@pyd.validate_call
def alias(**aliases: str | list[str]) -> t.Callable:
    """Alias command options.

    Decorates a function by attaching command option alias metadata,
    to be used at compile time to alias :py:class:`click.Option` objects.

    Aliases may only be defined for command-line options, not arguments.
    This translates to keyword-only parameters, i.e. those
    positioned after the ``*`` operator in a function signature.

    Parameters
    ----------
    **aliases:
        Mapping of option names to aliases.
        Option names must be keyword-only parameters in the decorated
        function signature.

    Returns
    -------
    Function decorated with command option alias metadata.

    Examples
    --------
    Aliasing a single option.

    >>> import feud
    >>> @feud.alias(verbose="-v")
    ... def func(*, verbose: bool) -> bool:
    ...     return verbose
    >>> feud.run(func, ["--verbose"], standalone_mode=False)
    True
    >>> feud.run(func, ["-v"], standalone_mode=False)
    True
    >>> feud.run(func, ["--no-verbose"], standalone_mode=False)
    False
    >>> feud.run(func, ["--no-v"], standalone_mode=False)
    False

    Aliasing a single option with multiple aliases.

    >>> import feud
    >>> @feud.alias(verbose=["-v", "-V"])
    >>> def func(*, verbose: bool) -> bool:
    ...     return verbose
    >>> feud.run(func, ["-v"], standalone_mode=False)
    True
    >>> feud.run(func, ["--no-v"], standalone_mode=False)
    False
    >>> feud.run(func, ["-V"], standalone_mode=False)
    True
    >>> feud.run(func, ["--no-V"], standalone_mode=False)
    False

    Aliasing multiple options.

    >>> import feud
    >>> from feud.typing import Counter
    >>> @feud.alias(verbose="-v", stringify="-s")
    ... def func(*, verbose: Counter, stringify: bool) -> int | str:
    ...     return f"Verbose level: {verbose}" if stringify else verbose
    >>> feud.run(func, ["-vvvs"], standalone_mode=False)
    "Verbose level: 3"
    >>> feud.run(func, ["-v", "-v", "-v", "-s"], standalone_mode=False)
    "Verbose level: 3"
    >>> feud.run(func, ["-vvv", "--no-s"], standalone_mode=False)
    3
    """

    def decorator(f: t.Callable) -> t.Callable:
        # check provided aliases and parameters match
        sig = inspect.signature(f)
        specified = set(aliases.keys())
        received = {
            p.name for p in sig.parameters.values() if p.kind == p.KEYWORD_ONLY
        }
        if len(specified - received) > 0:
            msg = (
                f"Arguments provided to 'alias' decorator must "
                f"also be keyword parameters for function {f.__name__!r}. "
                f"Received extra arguments: {specified - received!r}."
            )
            raise CompilationError(msg)

        # check alias uniqueness
        flat_aliases: list[str] = []
        for alias in aliases.values():
            if isinstance(alias, str):
                flat_aliases.append(alias)
            else:
                flat_aliases.extend(alias)

        if len(set(flat_aliases)) < len(flat_aliases):
            msg = f"Aliases for function {f.__name__!r} must be unique."
            raise CompilationError(msg)

        # check alias formats
        fmt = r"^-[a-zA-Z0-9]$"
        invalid = [alias for alias in flat_aliases if not re.match(fmt, alias)]
        if invalid:
            msg = (
                f"Invalid aliases provided for function {f.__name__!r}. "
                f"Aliases must be of the format {fmt!r}."
            )
            raise CompilationError(msg)

        f.__feud_aliases__ = {
            k: ([v] if isinstance(v, str) else v) for k, v in aliases.items()
        }
        return f

    return decorator


def env(**envs: str) -> t.Callable:
    """Specify environment variable inputs for command options.

    Decorates a function by attaching command option environment variable
    metadata, to be used at compile time by :py:class:`click.Option` objects.

    Environment variables may only be defined for command-line options, not
    arguments. This translates to keyword-only parameters, i.e. those
    positioned after the ``*`` operator in a function signature.

    Parameters
    ----------
    **envs:
        Mapping of option names to environment variables.
        Option names must be keyword-only parameters in the decorated
        function signature.

    Returns
    -------
    Function decorated with command option environment variable metadata.

    Examples
    --------
    Using an environment variable for a single option.

    >>> import os
    >>> import feud
    >>> @feud.env(token="TOKEN")
    ... def func(*, token: str) -> str:
    ...     return token
    >>> os.environ["TOKEN"] = "Hello world!"
    >>> feud.run(func, [], standalone_mode=False)
    "Hello World!"

    Using environment variables for multiple options.

    >>> import os
    >>> import feud
    >>> @feud.env(token="TOKEN", key="API_KEY")
    >>> def func(*, token: str, key: str) -> tuple[str, str]:
    ...     return token, key
    >>> os.environ["TOKEN"] = "Hello world!"
    >>> os.environ["API_KEY"] = "This is a secret key."
    >>> feud.run(func, [], standalone_mode=False)
    ("Hello world!", "This is a secret key.")
    """

    def decorator(f: t.Callable) -> t.Callable:
        # check provided envs and parameters match
        sig = inspect.signature(f)
        specified = set(envs.keys())
        received = {
            p.name for p in sig.parameters.values() if p.kind == p.KEYWORD_ONLY
        }
        if len(specified - received) > 0:
            msg = (
                f"Arguments provided to 'env' decorator must "
                f"also be keyword parameters for function {f.__name__!r}. "
                f"Received extra arguments: {specified - received!r}."
            )
            raise CompilationError(msg)

        f.__feud_envs__ = envs
        return f

    return decorator


def rename(command: str | None = None, /, **params: str) -> t.Callable:
    """Rename a command and/or its parameters.

    Useful for command/parameter names that use hyphens, reserved Python
    keywords or in-built function names.

    Parameters
    ----------
    command:
        New command name. If ``None``, the command is not renamed.

    **params:
        Mapping of parameter names to new names.
        Parameter names must be defined in the decorated function signature.

    Returns
    -------
    Function decorated with command/parameter renaming metadata.

    Examples
    --------
    Renaming a command.

    >>> import feud
    >>> @feud.rename("my-func")
    ... def my_func(arg_1: int, *, opt_1: str, opt_2: bool):
    ...     pass

    Renaming parameters.

    >>> import feud
    >>> @feud.rename(arg_1="arg-1", opt_2="opt-2")
    ... def my_func(arg_1: int, *, opt_1: str, opt_2: bool):
    ...     pass

    Renaming a command and parameters.

    >>> import feud
    >>> @feud.rename("my-func", arg_1="arg-1", opt_2="opt-2")
    ... def my_func(arg_1: int, *, opt_1: str, opt_2: bool):
    ...     pass
    """

    def decorator(f: t.Callable) -> t.Callable:
        # check provided names and parameters match
        sig = inspect.signature(f)
        specified = set(params.keys())
        received = {p.name for p in sig.parameters.values()}
        if len(specified - received) > 0:
            msg = (
                f"Arguments provided to 'env' decorator must "
                f"also be parameters for function {f.__name__!r}. "
                f"Received extra arguments: {specified - received!r}."
            )
            raise CompilationError(msg)

        f.__feud_names__ = _command.NameDict(command=command, params=params)
        return f

    return decorator



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import inspect

import pytest

from feud import click
from feud import typing as t
from feud._internal import _command


def test_get_context_ctx_no_annotation() -> None:
    def f(ctx, a: t.Any, b: t.Any, c: t.Any) -> None:  # noqa: ANN001
        pass

    sig = inspect.signature(f)
    assert _command.pass_context(sig)


def test_get_context_ctx_with_annotation() -> None:
    def f(ctx: click.Context, a: t.Any, b: t.Any, c: t.Any) -> None:
        pass

    sig = inspect.signature(f)
    assert _command.pass_context(sig)


def test_get_context_wrong_name_no_annotation() -> None:
    def f(context, a: t.Any, b: t.Any, c: t.Any) -> None:  # noqa: ANN001
        pass

    sig = inspect.signature(f)
    assert not _command.pass_context(sig)


def test_get_context_wrong_name_with_annotation() -> None:
    def f(context: click.Context, a: t.Any, b: t.Any, c: t.Any) -> None:
        pass

    sig = inspect.signature(f)
    assert not _command.pass_context(sig)


def test_get_context_wrong_position() -> None:
    def f(a: t.Any, b: t.Any, c: t.Any, ctx: click.Context) -> None:
        pass

    sig = inspect.signature(f)
    assert not _command.pass_context(sig)


def test_get_context_positional_only() -> None:
    def f(ctx: click.Context, /, a: t.Any, b: t.Any, c: t.Any) -> None:
        pass

    sig = inspect.signature(f)
    assert _command.pass_context(sig)


def test_get_context_keyword_only() -> None:
    def f(*, ctx: click.Context, a: t.Any, b: t.Any, c: t.Any) -> None:
        pass

    sig = inspect.signature(f)
    assert _command.pass_context(sig)


@pytest.mark.parametrize(
    ("hint", "negate_flags", "expected"),
    [
        (str, False, "--opt"),
        (str, True, "--opt"),
        (bool, False, "--opt"),
        (bool, True, "--opt/--no-opt"),
    ],
)
def test_get_option(hint: type, *, negate_flags: bool, expected: str) -> None:
    assert (
        _command.get_option("opt", hint=hint, negate_flags=negate_flags)
        == expected
    )


@pytest.mark.parametrize(
    ("hint", "negate_flags", "expected"),
    [
        (str, False, "-o"),
        (str, True, "-o"),
        (bool, False, "-o"),
        (bool, True, "-o/--no-o"),
    ],
)
def test_get_alias(hint: type, *, negate_flags: bool, expected: str) -> None:
    assert (
        _command.get_alias("-o", hint=hint, negate_flags=negate_flags)
        == expected
    )


@pytest.mark.parametrize(
    ("click_kwargs", "name", "expected"),
    [
        (
            {"commands": ["a", "b", "c"], "key": "value"},
            "test",
            {"key": "value", "name": "test"},
        ),
        (
            {"key": "value"},
            "a_name",
            {"key": "value", "name": "a_name"},
        ),
        (
            {"key": "value"},
            "_a_name",
            {"key": "value", "name": "_a_name"},
        ),
        (
            {"key": "value"},
            "-a-name",
            {"key": "value", "name": "a-name"},
        ),
        (
            {"key": "value"},
            "--a-name",
            {"key": "value", "name": "a-name"},
        ),
        (
            {"key": "value"},
            "AName",
            {"key": "value", "name": "aname"},
        ),
        (
            {"key": "value"},
            "ClassName",
            {"key": "value", "name": "classname"},
        ),
    ],
)
def test_sanitize_click_kwargs(
    click_kwargs: dict[str, str], name: str, expected: dict[str, str]
) -> None:
    _command.sanitize_click_kwargs(click_kwargs, name=name)
    assert click_kwargs == expected



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Overrides for ``click``."""

try:
    from rich_click import *
except ImportError:
    from click import *

from feud.click.context import *



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import re

import feud
from feud.version import VERSION, version_info


def test_version() -> None:
    """Check that the version is a valid SemVer version."""
    assert re.match(r"\d+\.\d+\.\d+[a-z0-9]*", VERSION)


def test_version_info() -> None:
    """Check that the version appears in the version info.

    FIXME: Not a thorough check of version_info() details.
    """
    assert VERSION in version_info()


def test_dunder() -> None:
    """Check that VERSION is the same as __version__."""
    assert feud.__version__ == VERSION



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Main invoke task collection."""

from __future__ import annotations

from invoke.collection import Collection
from invoke.config import Config
from invoke.tasks import task

from make import cov, docs, lint, release, tests


@task
def install(c: Config) -> None:
    """Install package with pre-commit hooks and core, dev, docs, & test
    dependencies.
    """
    # install dependencies
    # NOTE: only including docs/tests dependencies to please editors
    c.run("poetry install --sync --only base,main,dev,docs,tests -E all")
    # install pre-commit hooks
    c.run("pre-commit install --install-hooks")


@task
def clean(c: Config) -> None:
    """Clean temporary files, local cache and build artifacts."""
    commands: list[str] = [
        "rm -rf `find . -name __pycache__`",
        "rm -f `find . -type f -name '*.py[co]'`",
        "rm -f `find . -type f -name '*~'`",
        "rm -f `find . -type f -name '.*~'`",
        "rm -rf .cache",
        "rm -rf .pytest_cache",
        "rm -rf .ruff_cache",
        "rm -rf .tox",
        "rm -rf htmlcov",
        "rm -rf *.egg-info",
        "rm -f .coverage",
        "rm -f .coverage.*",
        "rm -rf build",
        "rm -rf dist",
        "rm -rf site",
        "rm -rf docs/build",
        "rm -rf coverage.xml",
    ]
    for command in commands:
        c.run(command)


# create top-level namespace
namespace = Collection()

# register top-level commands
for t in (install, clean):
    namespace.add_task(t)

# register namespaces
for module in (docs, tests, cov, lint, release):
    collection = Collection.from_module(module)
    namespace.add_collection(collection)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Custom types for counting options."""

from __future__ import annotations

import typing as t

from pydantic import conint

__all__ = ["Counter", "concounter"]


class CounterType:
    pass


#: Keep count of a repeated command-line option.
#:
#: Examples
#: --------
#: >>> import feud
#: >>> from feud.typing import Counter
#: >>> @feud.alias(verbose="-v")
#: ... def func(*, verbose: Counter) -> int:
#: ...     return verbose
#: >>> feud.run(func, ["-vvv"], standalone_mode=False)
#: 3
Counter = t.Annotated[int, CounterType]


def concounter(
    *,
    strict: bool | None = None,
    gt: int | None = None,
    ge: int | None = None,
    lt: int | None = None,
    le: int | None = None,
    multiple_of: int | None = None,
) -> t.Annotated[int, ...]:
    """Wrap :py:obj:`pydantic.types.conint` to allow for constrained counting
    options.

    Parameters
    ----------
    strict:
        Whether to validate the integer in strict mode. Defaults to ``None``.

    gt:
        The value must be greater than this.

    ge:
        The value must be greater than or equal to this.

    lt:
        The value must be less than this.

    le:
        The value must be less than or equal to this.

    multiple_of:
        The value must be a multiple of this.

    Returns
    -------
    The wrapped :py:obj:`pydantic.types.conint` type.

    Examples
    --------
    >>> import feud
    >>> from feud.typing import concounter
    >>> @feud.alias(verbose="-v")
    ... def func(*, verbose: concounter(le=3)) -> int:
    ...     return verbose
    >>> feud.run(func, ["-vvv"], standalone_mode=False)
    3
    """
    return t.Annotated[
        conint(
            strict=strict,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
        ),
        CounterType,
    ]


def is_counter(hint: t.Any) -> bool:
    args = t.get_args(hint)
    if len(args) > 1:
        return args[0] is int and CounterType in args
    return False



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Officially supported types from the ``typing`` package."""

from typing import (
    Annotated,
    Any,
    Deque,
    FrozenSet,
    List,
    Literal,
    NamedTuple,
    Optional,
    Pattern,
    Set,
    Text,
    Tuple,
    Union,
)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import click

import feud


@feud.command
def func(ctx: feud.click.Context) -> None:
    return ctx


def test_no_validation() -> None:
    """Test that the context argument is not validated,
    and returns the current Click context.
    """
    ctx: click.core.Context = func([], standalone_mode=False)
    assert isinstance(ctx, click.core.Context)
